/* 
   S3 964 Chipset-Driver - for SPEA Mercury PCI 2MB VRAM on 16 Apr 1995

   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/kernel.h>

/* THIS IS FOR TESTING PURPOSES ! COMMENT OUT FOR RELEASES ! */
#undef suser
#define suser() 1

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/malloc.h>
#include <linux/fcntl.h>
#include <linux/vt.h>
#include <sys/ioctl.h>
#include <asm/io.h>
#include <asm/pgtable.h>
#include "graphdev.h"
#include "kgi.h"
#include "vga_prg.h"
#include "vga_generic.h"
#include "generic.h"
#include "S3/regs3.h"
#include "S3/s3.h"

int set16col(void)
{ Outb_GR_CTL(4,(*(int *)parameters)& 3);	/* Read Map select */
  Outb_GR_CTL(0,(*(int *)parameters)&15);	/* Color select */
  return(0);
}

/*********************************************
**** Accelerated drawing functions follow ****
*********************************************/

int s3_wait_fifo(int spc_need)
{ int x=1000000000;
  while((inb(GP_STAT)&(0x100>>spc_need))&&--x);
  return(!x);
}

static int s3_wait_idle(void)
{ int x=1000000000;
  while((inw(GP_STAT)&0x200)&&--x);
  return(!x);
}

static int s3_drawbox(void)
{
  if (CurrState->bgmode||
      (CurrState->ModeNow.trflags&TRF_NO_ACCEL))
    return(-ENODRVSUP_ALWAYS_MMAP);

  if (*((int *)parameters+3)&&*((int *)parameters+2))	/* 0 width is not OK ! */
  {
	outl(get_fs_long(&CurrState->drawinfo->fg_color),FRGD_COLOR);

	if (s3_wait_fifo(6)) return(-ENODRVSUP_NOW_MMAP);
	outw(0xa000,			MULTIFUNC_CNTL);
	outw(*((int *)parameters  ),	CUR_X);
	outw(*((int *)parameters+1),	CUR_Y);
	outw((*((int *)parameters+2)-1),			MAJ_AXIS_PCNT);
	outw((*((int *)parameters+3)-1) | MIN_AXIS_PCNT,	MULTIFUNC_CNTL);
	outw(CMD_RECT | INC_Y | INC_X | DRAW | PLANAR | WRTDATA,CMD);
	s3_wait_idle();
  }
  return(0);
}

static int s3_drawline(void)
{ int dx,dy,min,max;
  if (CurrState->bgmode||
      (CurrState->ModeNow.trflags&TRF_NO_ACCEL))   
    return(-ENODRVSUP_ALWAYS_MMAP);

  dx=*((int *)parameters+2)-*((int *)parameters+0);
  dy=*((int *)parameters+3)-*((int *)parameters+1);
  
  if (abs(dx)>abs(dy))	{max=abs(dx);min=abs(dy);}
  else 			{max=abs(dy);min=abs(dx);}

  if (max==0) return(0);	/* FIXME ! */

  if (s3_wait_fifo(2)) return(-ENODRVSUP_NOW_MMAP);

  outl(get_fs_long(&CurrState->drawinfo->fg_color),FRGD_COLOR);
  outw(0xa000,			MULTIFUNC_CNTL);

  if (s3_wait_fifo(7)) return(-ENODRVSUP_NOW_MMAP);
  outw(*((int *)parameters  ),	CUR_X);
  outw(*((int *)parameters+1),	CUR_Y);

  outw((max-1),			MAJ_AXIS_PCNT);
  outw((min-max)<<1,		DESTX_DIASTP);
  outw((min<<1),		DESTY_AXSTP);

  outw((min+min-max-(dx<0)),	ERR_TERM);

  outw(CMD_LINE | ((dy>0)            ? INC_Y : 0) 
  		| ((dx>0)            ? INC_X : 0)
  		| ((abs(dy)>abs(dx)) ? YMAJAXIS : 0)
  		| DRAW | PLANAR | WRTDATA,CMD);

  s3_wait_idle();
  return(0);
}

/*****************************************************************************
****  Graphics  Device  Interface  Funtions                               ****
*****************************************************************************/

int kgi_graphics_ioctl(struct inode *inode, struct file *file, \
        unsigned int cmd, unsigned long arg)
{ 
	DEV_NUM_SETUP;

	if (CurrState->bgmode)			/* Don't accel inactive screen ! */
		return(-ENODRVSUP_MODE_MMAP);

	/* Misc stuff like setting the drawing color in 16-col VGA mode -
	   this is no Accel command and thus can be done even if NO_ACCEL
	   is set. */ 
	if (cmd==IO_GRAPH_SET16COL)
		{ COPYPAR(sizeof(int));
        	  return(set16col());  }


  /* If the NoAccel Flag is set, we cannot do acceleration for some reason
     - suggest using MMAP for this mode. */

	if (CurrState->ModeNow.trflags&TRF_NO_ACCEL)   
		return(-ENODRVSUP_MODE_MMAP);

  switch(cmd) { 

	/* No Accel support for those functions */

	case IO_GRAPH_DRAWPIXEL   : /* Faster than doing it here */
	case IO_GRAPH_GETPIXEL    :
	case IO_GRAPH_PUTPIXEL    :
	case IO_GRAPH_DRAWHLINE   :
	case IO_GRAPH_GETHLINE    :
	case IO_GRAPH_PUTHLINE    :
	case IO_GRAPH_DRAWVLINE   :
	case IO_GRAPH_GETVLINE    :
	case IO_GRAPH_PUTVLINE    :
	case IO_GRAPH_GETBOX      :
	case IO_GRAPH_PUTBOX      :return(-ENODRVSUP_ALWAYS_MMAP);

	/* Really accelerated functions */

        case IO_GRAPH_DRAWBOX  :COPYPAR(4*sizeof(int)+sizeof(void *));
    				return(s3_drawbox());
        case IO_GRAPH_DRAWLINE :COPYPAR(4*sizeof(int)+sizeof(void *));
    				return(s3_drawline());

	/* Functions that will probably benefit from lower-lever accel */

        case IO_GRAPH_FILLSCREEN:
        default		        :return -ENODRVSUP_ALWAYS_LOWER;
   }
   /* We should never get here ... */
   return -ENODRVSUP_ALWAYS_CANT;
}                                                      
