/* 
   S3 Dumpreg-Driver - will dump regs and terminate when insmodded.

   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/malloc.h>
#include <linux/fcntl.h>
#include <linux/vt.h>
#include <sys/ioctl.h>
#include <asm/io.h>
#include <asm/pgtable.h>
#include "graphdev.h"
#include "kgi-module.h"
#include "vga_prg.h"
#include "spea_mercury.h"
#include "vga_generic.h"
#include "generic.h"
#include "S3/regs3.h"
#include "S3/s3.h"

char kernel_version[]= UTS_RELEASE;

/*****************************************************************************
*** Chipset characteristics                                                ***
*****************************************************************************/

struct ggi_ChipParams ChipPar = 
	{	2*1024*1024,		/* 2MB RAM */
		2048,
		1024,
		"GENERIC S3",
		"Dumpreg",
		0,	
		/* more is not implemented now ... */
	};

/*****************************************************************************
*** Enable S3 extended register access                                     ***
*****************************************************************************/
void unlock_s3(void)
{ 
	Outb_CRTC(CRT_LOCK1, CR38_MAGIC);	/* unlock CR30-CR3F */
	Outb_CRTC(CRT_LOCK2, CR39_MAGIC);	/* unlock CR40-CRFF */
	Outb_CRTC(CRT_LOCK,Inb_CRTC(CRT_LOCK)&0xcf);	/* unlock H/VTiming */
	Outb_SEQ (SEQ_UNLOCK, SR8_MAGIC);	/* unlock VESAPower */
}

void lock_s3(void)
{ 
	Outb_CRTC(CRT_LOCK1 , 0);		/* lock CR30-CR3F */
	Outb_CRTC(CRT_LOCK2 , 0);		/* lock CR40-CRFF */
	Outb_SEQ (SEQ_UNLOCK, 0);		/* lock VESAPower */
}

/*****************************************
*** Check if this is a vision 964 chip ***
*****************************************/
static int Detect_Vision964(void)
{
	register unsigned char cr36;

	cr36 = Inb_CRTC(CRT_CHIPID);
	if ( (cr36 & CR30_CHIPID) != 0xD0)
	{
		printk("  Vision964 not detected (ID = %.2x).\n", cr36 & CR30_CHIPID);
		return 0;
	}
	printk("  Vision964 rev %i detected.\n", cr36 & CR30_REVISION);

	cr36 = Inb_CRTC(CRT_CONFIG1);
	switch(cr36 & CR36_MEMSIZE_MASK)
	{
		case CR36_1MB: 
			ChipPar.ramsize = 0x100000;break;
		case CR36_2MB:
			ChipPar.ramsize = 0x200000;break;
		case CR36_3MB: 
			ChipPar.ramsize = 0x300000;break;
		case CR36_4MB:
			ChipPar.ramsize = 0x400000;break;
		case CR36_6MB: 
			ChipPar.ramsize = 0x600000;break;
		case CR36_8MB:
			ChipPar.ramsize = 0x800000;break;
		default:
			printk("%s: can't determine VRAM size, assume 1MB.\n",__FILE__);
	}

	switch(cr36 & CR36_RAMMASK)
	{
		case CR36_FASTPAGERAM:
			printk("  %i FP RAM, ",  ChipPar.ramsize); break;
		case CR36_EDO_RAM:
			printk("  %i EDO RAM, ", ChipPar.ramsize); break;
		default:
			printk("  %i unknown RAM, ", ChipPar.ramsize);
	}

	switch(cr36 & CR36_BUSMASK)
	{
		case CR36_PCI      :
			printk("PCI bus\n");        break;
		case CR36_VESALOCAL:
			printk("Vesa Local bus\n"); break;
		default:
			printk("unknown bus\n");
     }

  return 1;
}

/****************************/
/*** Set start of display ***/
/****************************/

void hw_setdispstart(int dx,int dy) {}

/*****************************************************************************
****  Graphics  Device  Interface  Funtions                               ****
*****************************************************************************/

int kgi_ChipsetIoctl(struct inode *inode, struct file *file, \
        unsigned int cmd, unsigned long arg)
{ return -ENODRVSUP_ALWAYS_CANT; }                                                      

void dumpit(char *buf,int anz,char *comment)
{ int x;
  printk("%s:\n{",comment);
  for(x=0;x<anz;x++)
  { printk("%2x, ",(int)((unsigned char)*buf++));
    if ((x&15)==15) printk("\n");
  }
  printk("}\n");
}

void dump_regs(void)
{ char buffer[128];

  SaveCRTC(buffer,128);  
  dumpit  (buffer,128,"CRTC");  

  SaveSEQ (buffer,5);  
  dumpit  (buffer,5,"Sequencer");  

  SaveATR (buffer,21);  
  dumpit  (buffer,21,"Attribute Controller");  

  SaveGR  (buffer,9);  
  dumpit  (buffer,9,"Graphics Controller");  
}

int kgi_ChipsetInit(void)
{
	outb(S3_DISABLE, S3_SUBSYSTEM);	/* chip wakeup */
	outb(S3_WAKEUP,  S3_SETUP);
	outb(S3_ENABLE,  S3_SUBSYSTEM);

	Detect_Vision964();		/* FIXME - should react to rc */
  
	unlock_s3();			/* Unlock s3-regs ... other Inits might need access */
  
	printk("Manufacturer : %s\n",ChipPar.manufact);
	printk("Model        : %s\n",ChipPar.model);
	printk("RAM / Maxsize: %d Bytes (%d x %d) \n",
        	ChipPar.ramsize,
        	ChipPar.xmaxdots,
        	ChipPar.ymaxdots);

	_getsetfont_hw(OpenFont,0);
	dump_regs();
	return 0;			/* This never goes OK. */
}

void kgi_ChipsetCleanup(void) 
{ lock_s3(); }

void kgi_GraphmapClose(struct vm_area_struct * area) {}
unsigned long kgi_GraphmapNopage(struct vm_area_struct * area, unsigned long address,
	unsigned long page, int no_share) {return 0;}
int kgi_CheckChipsetTiming(struct ggi_Timing *TM,int olderr) {return 0;}
int kgi_SetChipsetTiming(struct ggi_Timing *TM) {return 0;}


void ramdac_select_RS23(int reg)
{  Outb_CRTC(CRT_DACCONTROL,(Inb_CRTC(CRT_DACCONTROL) &
				0xFC) | ((reg & 0x0C) >> 2));
}
