/* 
   S3 964 Chipset-Driver - for SPEA Mercury PCI 2MB VRAM on 16 Apr 1995

   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/kernel.h>

/* THIS IS FOR TESTING PURPOSES ! COMMENT OUT FOR RELEASES ! */
#undef suser
#define suser() 1

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/malloc.h>
#include <linux/fcntl.h>
#include <linux/vt.h>
#include <sys/ioctl.h>
#include <asm/io.h>
#include <asm/pgtable.h>
#include "graphdev.h"
#include "kgi.h"
#include "vga_prg.h"
#include "spea_mercury.h"
#include "vga_generic.h"
#include "generic.h"
#include "S3/regs3.h"
#include "S3/s3.h"

char kernel_version[]= UTS_RELEASE;

/*****************************************************************************
*** Chipset characteristics                                                ***
*****************************************************************************/

struct kgi_chipset chip_par = 
	{
		"SPEA / Video 7",
		"Mercury PCI S3-964",
		0,			/* Modes  - dep. */
		FL_CHIP_LMAP
		| FL_CHIP_DOUBLESCAN,
		{ 2048, 1024} ,
		1024*1024,		/* 1MB RAM assumed */
		{ 0, 135000000 },
	};

/*****************************************************************************
*** Enable S3 extended register access                                     ***
*****************************************************************************/
static void unlock_s3(void)
{ 
	Outb_CRTC(CRT_LOCK1, CR38_MAGIC);	/* unlock CR30-CR3F */
	Outb_CRTC(CRT_LOCK2, CR39_MAGIC);	/* unlock CR40-CRFF */
	Outb_CRTC(CRT_LOCK,Inb_CRTC(CRT_LOCK)&0xcf);	/* unlock H/VTiming */
	Outb_SEQ (SEQ_UNLOCK, SR8_MAGIC);	/* unlock VESAPower */
}

static void lock_s3(void)
{ 
	Outb_CRTC(CRT_LOCK1 , 0);		/* lock CR30-CR3F */
	Outb_CRTC(CRT_LOCK2 , 0);		/* lock CR40-CRFF */
	Outb_SEQ (SEQ_UNLOCK, 0);		/* lock VESAPower */
}

/****************************************************************************
*** Generic ramdac and clock communication                                ***
****************************************************************************/

static inline void setrs23(unsigned char dac_reg)
{	Outb_CRTC(CRT_DACCONTROL,
			(Inb_CRTC(CRT_DACCONTROL)&~CR55_DAC_MASK) 
			|(dac_reg >> 2));
}

void outbDAC(unsigned char value, unsigned char dac_reg)
{   	setrs23(dac_reg);outb(value, 0x3c6+(dac_reg&0x03));setrs23(0); }

unsigned char
inbDAC(unsigned char dac_reg)
{ 	unsigned char rc;
	setrs23(dac_reg);
   	rc=inb(0x3c6+(dac_reg&0x03));
   	setrs23(0);
   	return rc;
}

void
insbDAC(unsigned short dac_reg, void *buf, int count)
{	setrs23(dac_reg);
	insb(0x3c6+(dac_reg & 0x03), buf, count);
	setrs23(0);
}

void
outsbDAC(unsigned short dac_reg, void *buf, int count)
{	setrs23(dac_reg);
	outsb(0x3c6+(dac_reg & 0x03), buf, count);
	setrs23(0);
}

void
setCLK(int value)
{	Outb_CRTC(CRT_EXTMODE1,(Inb_CRTC(CRT_EXTMODE1) & ~CR42_DCLK_MASK) 
			       | value);
}

int
getCLK(void)
{
  return Inb_CRTC(CRT_EXTMODE1) & CR42_DCLK_MASK;
}



/*****************************************
*** Check if this is a vision 964 chip ***
*****************************************/
static int Detect_Vision964(void)
{
	register unsigned char cr36;
	
	unlock_s3();

	cr36 = Inb_CRTC(CRT_CHIPID);
	if ( (cr36 & CR30_CHIPID) != 0xD0)
	{
		printk("  Vision964 not detected (ID = %.2x).\n", cr36 & CR30_CHIPID);
		return 0;
	}
	printk("  Vision964 rev %i detected.\n", cr36 & CR30_REVISION);

	cr36 = Inb_CRTC(CRT_CONFIG1);
	switch(cr36 & CR36_MEMSIZE_MASK)
	{
		case CR36_1MB: 
			chip_par.ramsize = 0x100000;break;
		case CR36_2MB:
			chip_par.ramsize = 0x200000;break;
		case CR36_3MB: 
			chip_par.ramsize = 0x300000;break;
		case CR36_4MB:
			chip_par.ramsize = 0x400000;break;
		case CR36_6MB: 
			chip_par.ramsize = 0x600000;break;
		case CR36_8MB:
			chip_par.ramsize = 0x800000;break;
		default:
			printk("%s: can't determine VRAM size, assume 1MB.\n",__FILE__);
	}

	switch(cr36 & CR36_RAMMASK)
	{
		case CR36_FASTPAGERAM:
			printk("  %i FP RAM, ",  chip_par.ramsize); break;
		case CR36_EDO_RAM:
			printk("  %i EDO RAM, ", chip_par.ramsize); break;
		default:
			printk("  %i unknown RAM, ", chip_par.ramsize);
	}

	switch(cr36 & CR36_BUSMASK)
	{
		case CR36_PCI      :
			printk("PCI bus\n");        break;
		case CR36_VESALOCAL:
			printk("Vesa Local bus\n"); break;
		default:
			printk("unknown bus\n");
     }

  return 1;
}

/**********************************************/
/* Set board to a defined state and unlock it */
/**********************************************/

void kgi_chipset_reset(void)
{	static char SEQ [5]=
		{ 0x03, 0x01, 0x03, 0x00, 0x02};
	static char GR  [9]=
		{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x0e, 0x00, 0xFF };
	static char ATR[21]=
		{ 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x14, 0x07,
  		  0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f,
		  0x0c, 0x00, 0x0f, 0x00, 0x00 };
	static char CRTC[128]=
		{ 0x82, 0x63, 0x69, 0x9d, 0x6a, 0x9d, 0x88, 0xf0,
		  0x00, 0x6f, 0x0e, 0x0f, 0x00, 0x00, 0x00, 0x64,
		  0x51, 0x22, 0x4f, 0x32, 0x1f, 0x51, 0x52, 0xa3,
 		  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
 		  0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	 	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 		  0xd0, 0x05, 0x00, 0x00, 0x00, 0x00, 0x9e, 0x0f,
 		  0x48, 0xa5, 0x05, 0x5a, 0x10, 0x10, 0x10, 0x10,
	 	  0x11, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff,
 		  0xfc, 0xff, 0xff, 0x00, 0x1f, 0xff, 0xff, 0xdf,
	 	  0x00, 0x00, 0x04, 0x00, 0x38, 0x00, 0x00, 0x00,
	 	  0x03, 0x00, 0x00, 0x00, 0x02, 0x00, 0x00, 0x00,
	 	  0x07, 0x80, 0xa1, 0x00, 0x00, 0x00, 0x00, 0x00,
	 	  0xad, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	 	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	 	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };

	unlock_CRTC();
	unlock_s3();
	RestoreSEQ (SEQ ,   5);		/* Reset all registers to a defined state */
	RestoreCRTC(CRTC, 128);
	RestoreGR  (GR  ,   9);
	RestoreATR (ATR ,  21);

	Outb_SEQ(0x0d,Inb_SEQ(0x0d)&0x0f);	/* No power saving */
	outb((inb(0x3CC) & 0xF3),0x3C2);	/* Use 25.175MHz clock */

	outw(0x0000,ADVFUNC_CNTL);	/* Disable 8514A part */
}

static void ResetCard(void)
{
	kgi_chipset_reset();
	kgi_ramdac_reset();	/* BT485 will somehow disable clock-changing occasionally,
				   when in MUX mode ... strange - so we do a reset here */
	kgi_clock_reset();
	kgi_monitor_reset();
}

static void SetExtendedTiming(struct kgi_mode *TM,int xvirtshift)
{
	int	startFIFOfetch;
	int	h;

	startFIFOfetch=((TM->x.total+TM->x.syncstart)>>4)-2;
	
	Outb_CRTC(CRT_FIFOSTART,startFIFOfetch&0xff);

	Outb_CRTC(CRT_SYSCONTROL2, 
		  (Inb_CRTC(CRT_SYSCONTROL2)&0xc0)|
		  ((TM->request.virtual.x>>(4+xvirtshift))&0x30));

	Outb_CRTC(19,TM->request.virtual.x>>xvirtshift);

	Outb_CRTC(CRT_EV_OVERFLOW,
	(((TM->y.total  -2)&0x400)>>10)|
	(((TM->y.width  -1)&0x400)>> 9)|
	(((TM->y.syncstart)&0x400)>> 8)|
	(((TM->y.syncstart)&0x400)>> 6)|0x40 );

	h = Inb_CRTC(CRT_EH_OVERFLOW)&0x80;	/* Leave BGT alone */
	h|= 	((((TM->x.total    >> 3) - 5) & 0x100) >> 8) |
		((((TM->x.width     >> 3) - 1) & 0x100) >> 7) |
		((((TM->x.syncstart >> 3) - 1) & 0x100) >> 6) |
		((((TM->x.syncstart >> 3)    ) & 0x100) >> 4) |
		((((startFIFOfetch     )    ) & 0x100) >> 2) ;

	if ((TM->x.syncend >> 3) - (TM->x.syncstart >> 3) > 32)
		h |= CR5D_HSYNCEND_B5;   /* add another 32 DCLKs to hsync pulse width */

	if ((TM->x.syncend >> 3) - (TM->x.syncstart >> 3) > 64)
		h |= CR5D_HBLANKEND_B6;   /* add another 64 DCLKs to blank pulse width */

	Outb_CRTC(CRT_EH_OVERFLOW, h);

	Outb_CRTC(CRT_INTERLACESTART,	/* Interlace mode frame offset */
		  TM->request.virtual.x>>xvirtshift); 
}

/*****************************************************************************
*** Check/Setup an accelerated mode.                                       ***
*****************************************************************************/
static int CheckAccelmode(struct kgi_mode *TM,int olderr)
{ 
       if (TM->request.virtual.x<= 640) TM->request.virtual.x= 640;
  else if (TM->request.virtual.x<= 800) TM->request.virtual.x= 800;
  else if (TM->request.virtual.x<=1024) TM->request.virtual.x=1024;
  else if (TM->request.virtual.x<=1152) TM->request.virtual.x=1152;
  else if (TM->request.virtual.x<=1280) TM->request.virtual.x=1280;
  else if (TM->request.virtual.x<=1600) TM->request.virtual.x=1600;
  else if (TM->request.virtual.x<=2048) TM->request.virtual.x=2048;
  else { TM->request.virtual.x=2048;return(-ENOMODESUP_CHIP); }

	if (TM->request.visible.x&0x07)
		TM->request.visible.x&=0xfffffff8;	/* This must be a multiple of 8 */

	TM->clock=86000000;
	TM->request.trflags&=~TRF_NO_ACCEL;
	return(0);
}  

static int SetAccelmode(struct kgi_mode *TM,int xvirtshift)
{ 
	int h,x;
	char SEQ [5]=
		{ 0x03, 0x01, 0x0f, 0x00, 0x0e };
	char GR  [9]=
		{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x05, 0x0F, 0xFF };
	char ATR[21]=
		{ 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
  		  0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
  		  0x41, 0x00, 0x0f, 0x00, 0x00};

	ResetCard();
	RestoreSEQ(SEQ,5);
	RestoreGR (GR , 9);
	RestoreATR(ATR,21);    

	kgi_clock_set_mode(TM);
	kgi_ramdac_set_mode(TM);

	setup_VGA_mode(TM);

	Outb_CRTC(20,0x00);
	Outb_CRTC(23,0xE3);

	SetExtendedTiming(TM,xvirtshift);

	if (TM->request.virtual.x==2048)
		Outb_CRTC(CRT_MEMCONFIG0,0x8f);
	else 
		Outb_CRTC(CRT_MEMCONFIG0,0x8d);

	Outb_CRTC(CRT_BACKCOMPAT1,0x00);
	Outb_CRTC(CRT_BACKCOMPAT2,CR33_NOBORDER); /* ??? */
	Outb_CRTC(CRT_BACKCOMPAT3,CR34_ENB_FIFO);
	Outb_CRTC(CRT_LOCK,0x00);
	Outb_CRTC(CRT_SYSCONFIG,Inb_CRTC(CRT_SYSCONFIG)|CR40_UNLOCK_ENH);

	h=Inb_CRTC(CRT_SYSCONTROL1)&0x3e;		/* Width select */

	switch(TM->request.virtual.x)
	{
		case  640:h|=CR50_640;break;
		case  800:h|=CR50_800;break;
		case 1152:h|=CR50_1152;break;
		case 1280:h|=CR50_1280;break;
		case 1600:h|=CR50_1600;break;
		case 1024:h|=CR50_1024;break;
		case 2048:h|=CR50_2048;break;
		default  :printk("Illegal virtual.x !\n");break;		/* OK */
	}

	Outb_CRTC(CRT_SYSCONTROL1,h);

#if 0	/* Register undocumented ... does it exist ? */
	Outb_CRTC(0x44,0);					/* OK */
#endif
#ifdef OLD_MN_PAR
	h = 255;
	{	int clock,mclk;
		clock = TM->clock * 1;
		mclk = 60000;  /* 130 MHz, limit for 964 */
		m = (int)((mclk/1000.0*.72+16.867)*89.736/(clock/1000.0+39)-21.1543);
		m -= 0; /* Madjust */
		if (m > 31) m = 31;
		else if (m < 0) {
		 m = 0;
		h = 16;
		}
	}
	Outb_CRTC(CRT_MEMCONTROL2,m<<3);	/* always 0x38 ??? */
	h -= /*Nadjust*/ 0;		
	if (h < 0) h = 0;
	else if (h > 255) h = 255;
	Outb_CRTC(CRT_MEMCONTROL3,h);	/* always 0x07 ??? */

	printk("M: %d,N: %d\n",m,h);
#else
	Outb_CRTC(CRT_MEMCONTROL2,15<<3);	
	Outb_CRTC(CRT_MEMCONTROL3,15);
	/* Give memory 1:1 to FiFo and others - never seen this fail and the XF
	   method always gives 0,16 for m,n ... */
#endif
	for(x=0;x<2048/64;x++)
	{ 
		Outb_CRTC(CRT_EXTCONTROL4,x);		/* Bank select */
		memset((void *)0xa0000,0,0x10000);  /* erase graphics RAM */
	}
	Outb_CRTC(CRT_EXTCONTROL4,0);
	Outb_CRTC(CRT_LAWCONTROL,CR58_SAM_256);	/* We DO NOT use the linear window right now. */
	/*  Outb_CRTC(CRT_LAW_POS_H,0x0a); */	/* for this would eliminate the paging which  */
	/*  Outb_CRTC(CRT_LAW_POS_L,0x0a); */	/* is desireable for a demo-driver.           */

	/* Reset the Accelerator Engine */

	outw(0x0001,ADVFUNC_CNTL);
	outw(0x9000,SUBSYS_CNTL);
	outw(0x5000,SUBSYS_CNTL); /* ? 0x4 ?*/
	inw(SUBSYS_STAT); /* ??? */
	outw(MEM_CNTL | VRTCFG_4 | HORCFG_8,MULTIFUNC_CNTL);
	outw(MULT_MISC |0x10 , MULTIFUNC_CNTL);

        /* Now set some basic accelerator registers */

	s3_wait_fifo(6);
	outw(FSS_FRGDCOL | MIX_SRC,FRGD_MIX);
	outw(BSS_BKGDCOL | MIX_SRC,BKGD_MIX);		/* Use the colour directly */
	outw(SCISSORS_T |        0,MULTIFUNC_CNTL);	/* Clip the virtual area*/
	outw(SCISSORS_L |        0,MULTIFUNC_CNTL);
	outw(SCISSORS_R | (TM->request.virtual.x-1),MULTIFUNC_CNTL);
	outw(SCISSORS_B | (TM->request.virtual.y-1),MULTIFUNC_CNTL);

	s3_wait_fifo(5);
	outw(~0,WRT_MASK);				/* Enable all planes */
	outw(~0,WRT_MASK);				/* Enable all planes */
	outw(PIX_CNTL | 0x0000,MULTIFUNC_CNTL);
      
        outl(0,FRGD_COLOR);

	return(0);
}  

/*****************************************************************************
*** Check/Setup a Text-Mode of the given size                              ***
*****************************************************************************/
static int CheckText(struct kgi_mode *TM,int command)
{ 
	int x;

	if (TM->request.textgrid.x==7)
	{	TM->request.textgrid.x=8;
		if ((x=vga_CheckText(TM))) return x;
		TM->request.textgrid.x=7;
		TM->realclockmul.mul=7;
		TM->realclockmul.div=8;
	}
	else if ((x=vga_CheckText(TM))) return x;

	TM->clock=65000000;/* Max for text */

	TM->x.width=TM->request.visible.x;
	TM->y.width=TM->request.visible.y;
  
	TM->ymagnify=1;
  
	/* We can do nothing against RAMDAC objections */
	if ((x=kgi_ramdac_check_mode(TM,CMD_NONE))) return(x); 

	/* We can do nothing against Monitor objections except asking for
	   a less perfect mode */
	if ((x=kgi_monitor_check_mode(TM,CMD_MON_PERFECT)))
		x=kgi_monitor_check_mode(TM,CMD_MON_SOME);

	return x;
}  

static int SetText(struct kgi_mode *TM)
{ 	
	ResetCard();
	kgi_ramdac_set_mode(TM);
	kgi_clock_set_mode(TM);
	vga_SetText(TM);
	SetExtendedTiming(TM,2);
	if (TM->request.textgrid.x==7)
		Outb_CRTC(CRT_BACKCOMPAT1,(Inb_CRTC(CRT_BACKCOMPAT1)&0xfc)|0x01);
#if 0	
	/* DO NOT ENABLE ! This simply doesn't work :-( */
	Outb_CRTC(CRT_MEMCONFIG0,Inb_CRTC(CRT_MEMCONFIG0)|CR31_HISPEEDTXT);
	Outb_CRTC(CRT_MISC1     ,Inb_CRTC(CRT_MISC1)     |CR3A_HISPEEDTXT);
#endif
	return(0);
}  

/*****************************************************************************
*** Check/Setup a 4bit graphics-mode of the given size                     ***
*****************************************************************************/
static int Check4bit(struct kgi_mode *TM,int olderr)
{ 
	int	x, clksave;

#if 0
	if (TM->request.virtual.x*TM->request.virtual.y/8>64*1024) 
	{
		TM->request.virtual.x=800;TM->request.virtual.y=600;
		return(-ENOMODESUP_CHIP);
	}
#endif

	if (TM->request.visible.x&0x07) 
		TM->request.visible.x&=0xfffffff8;	/* This must be a multiple of 8 */

	TM->clock=86000000;	/* FIXME ... */
	TM->ymagnify=1;

	/* We can do nothing against RAMDAC objections */
	if ((x=kgi_ramdac_check_mode(TM,CMD_NONE))) return(x); 

	clksave=TM->clock;	/* RAMDAC might have limited it */

	while(1)
	{
		TM->x.width=TM->request.visible.x;
		TM->y.width=TM->request.visible.y*TM->ymagnify;

		switch((x=kgi_monitor_check_mode(TM,CMD_MON_PERFECT)))
		{ 
			case MR_Y_TOO_SMALL:
				TM->clock=clksave;
				if (TM->ymagnify++<32) break;	/* Retry with more magn. */
    			  	TM->ymagnify=1;
	    			x=kgi_monitor_check_mode(TM,CMD_MON_SOME);
    			  	return x;
			default:
				return x;	/* We can do nothing about it */
		} 
	}  
	
	return(-ENOMODESUP_CHIP);
}  

static int Set4bit(struct kgi_mode *TM)
{ 
	TM->muxmode=DAC_MUX_NORMAL;/* FIXME ! */
	ResetCard();
	kgi_ramdac_set_mode(TM);
	kgi_clock_set_mode(TM);
	vga_Set4bit(TM);
	SetExtendedTiming(TM,4);

	return(0);
}  

/*****************************************************************************
*** Check/Setup a 8bit graphics-mode of the given size                     ***
*****************************************************************************/
static int Check8bit(struct kgi_mode *TM,int olderr)
{ 
	int x, clksave;

	if ((x=CheckAccelmode(TM,olderr))) return(x);

	TM->ymagnify=1;

	/* We can do nothing against RAMDAC objections */
	if ((x=kgi_ramdac_check_mode(TM,CMD_NONE))) 
		return(x); 

	clksave=TM->clock;	/* RAMDAC might have limited it */

	while(1)
	{
		TM->x.width=TM->request.visible.x;
		TM->y.width=TM->request.visible.y*TM->ymagnify;

		switch((x=kgi_monitor_check_mode(TM,CMD_MON_PERFECT)))
		{ 
			case MR_Y_TOO_SMALL:
				TM->clock=clksave;
    				if (TM->ymagnify++<32) break;	/* Retry with more magn. */
   				TM->ymagnify=1;
    				x=kgi_monitor_check_mode(TM,CMD_MON_SOME);
    				return x;
			default:return x;	/* We can do nothing about it */
		} 
	}  

	return(-ENOMODESUP_CHIP);
}  

static int Set8bit(struct kgi_mode *TM)
{ 
	int h;
	SetAccelmode(TM,3);

	Outb_CRTC(CRT_MISC1,CR3A_NO4BPPMODE|
			    CR3A_ALT_REFRESH|
			    CR3A_REFRESH1  /* was 0xb5 */);

	switch(8)	/* bits per pixel ??? Documentation disapproves ! */
	{
		case  8:Outb_CRTC(CRT_EXTMODE2,0x0);break;
		case 16:
		case 32:Outb_CRTC(CRT_EXTMODE2,0x08);break;
	}

	Outb_CRTC(0x42,Inb_CRTC(0x42)|4); /* Spea Mercury peculiarity ... */

	switch(8)
	{	
		case 32:Outb_CRTC(CRT_HWC_MODE,(Inb_CRTC(CRT_HWC_MODE)&0xf2)|0x04);break;
		default:Outb_CRTC(CRT_HWC_MODE,Inb_CRTC(CRT_HWC_MODE)&0xf2);break;
	}							/* SVS3 says 0 ... */

	Outb_CRTC(CRT_SYSCONTROL1,Inb_CRTC(CRT_SYSCONTROL1)&0xcf);	/* 8 BPP */

	Outb_CRTC(CRT_DACCONTROL,/* was 0x68*/(Inb_CRTC(CRT_DACCONTROL)&0x08)|0x40);

	h=(TM->x.width*1)/4+1; /* FIXME? - other bbps, s3drv says 8 ! */
	Outb_CRTC(CRT_MEMCONTROL4,CR61_L_ENABLE|(h>>8));
	Outb_CRTC(CRT_MEMCONTROL5,(h&0xff));


	Outb_CRTC(CRT_EXTMISC0,0x20);
	Outb_CRTC(CRT_EXTMISC1,0x12);

#if 0
	Outb_CRTC(CRT_EXTMISC1,Inb_CRTC(CRT_EXTMISC1)|CR66_PIXELBUS_OFF);
	Outb_CRTC(CRT_EXTSYNC1,Inb_CRTC(CRT_EXTSYNC1)|CR56_SYNC_OFF);
#endif
	
	return(0);
}  

/*****************************************************************************
*** Check/Setup a 16 bit graphics-mode of the given size                   ***
*****************************************************************************/
static int Check16bit(struct kgi_mode *TM,int olderr)
{ 
	int x, clksave;

	if ((x=CheckAccelmode(TM,olderr))) return(x);

	TM->ymagnify=1;

	/* We can do nothing against RAMDAC objections */
	if ((x=kgi_ramdac_check_mode(TM,CMD_NONE))) 
		return(x); 

	clksave=TM->clock;	/* RAMDAC might have limited it */

	while(1)
	{
		TM->x.width=TM->request.visible.x;
		TM->y.width=TM->request.visible.y*TM->ymagnify;

		switch((x=kgi_monitor_check_mode(TM,CMD_MON_PERFECT)))
		{ 
			case MR_Y_TOO_SMALL:
				TM->clock=clksave;
    				if (TM->ymagnify++<32) break;	/* Retry with more magn. */
   				TM->ymagnify=1;
    				x=kgi_monitor_check_mode(TM,CMD_MON_SOME);
    				return x;
			default:return x;	/* We can do nothing about it */
		} 
	}  

	return(-ENOMODESUP_CHIP);
}  

static int Set16bit(struct kgi_mode *TM)
{ 
	int h;
	SetAccelmode(TM,2);

	Outb_CRTC(CRT_MISC1,CR3A_NO4BPPMODE|
			    CR3A_ALT_REFRESH|
			    CR3A_REFRESH1  /* was 0xb5 */);

	Outb_CRTC(CRT_EXTMODE2,0x08);

	Outb_CRTC(0x42,Inb_CRTC(0x42)|4); /* Spea Mercury peculiarity ... */

	switch(16)
	{	
		case 32:Outb_CRTC(CRT_HWC_MODE,(Inb_CRTC(CRT_HWC_MODE)&0xf2)|0x04);break;
		default:Outb_CRTC(CRT_HWC_MODE,Inb_CRTC(CRT_HWC_MODE)&0xf2);break;
	}							/* SVS3 says 0 ... */

	Outb_CRTC(CRT_SYSCONTROL1,
		  (Inb_CRTC(CRT_SYSCONTROL1)&0xcf)
		  |CR50_16BPP);			/* bits per pixel*/

	Outb_CRTC(CRT_DACCONTROL,/* was 0x68*/(Inb_CRTC(CRT_DACCONTROL)&0x08)|0x40);


	h=(TM->x.width*1)/4+1; /* FIXME? - other bbps, s3drv says 8 ! */
	Outb_CRTC(CRT_MEMCONTROL4,CR61_L_ENABLE|(h>>8));
	Outb_CRTC(CRT_MEMCONTROL5,(h&0xff));


	Outb_CRTC(CRT_EXTMISC0,0x20);
	Outb_CRTC(CRT_EXTMISC1,0x11);

	return(0);
}  

/*****************************************************************************
*** Check/Setup a 32(24) bit graphics-mode of the given size               ***
*****************************************************************************/
static int Check32bit(struct kgi_mode *TM,int olderr)
{ 
	int x, clksave;

	if ((x=CheckAccelmode(TM,olderr))) return(x);

	TM->ymagnify=1;

	/* We can do nothing against RAMDAC objections */
	if ((x=kgi_ramdac_check_mode(TM,CMD_NONE))) 
		return(x); 

	clksave=TM->clock;	/* RAMDAC might have limited it */

	while(1)
	{
		TM->x.width=TM->request.visible.x;
		TM->y.width=TM->request.visible.y*TM->ymagnify;

		switch((x=kgi_monitor_check_mode(TM,CMD_MON_PERFECT)))
		{ 
			case MR_Y_TOO_SMALL:
				TM->clock=clksave;
    				if (TM->ymagnify++<32) break;	/* Retry with more magn. */
   				TM->ymagnify=1;
    				x=kgi_monitor_check_mode(TM,CMD_MON_SOME);
    				return x;
			default:return x;	/* We can do nothing about it */
		} 
	}  

	return(-ENOMODESUP_CHIP);
}  

static int Set32bit(struct kgi_mode *TM)
{ 
	int h;
	SetAccelmode(TM,1);

	Outb_CRTC(CRT_MISC1,CR3A_NO4BPPMODE|
			    CR3A_ALT_REFRESH|
			    CR3A_REFRESH1  /* was 0xb5 */);

	Outb_CRTC(CRT_EXTMODE2,0x08);

	Outb_CRTC(0x42,Inb_CRTC(0x42)|4); /* Spea Mercury peculiarity ... */

	switch(32)
	{	
		case 32:Outb_CRTC(CRT_HWC_MODE,(Inb_CRTC(CRT_HWC_MODE)&0xf2)|0x04);break;
		default:Outb_CRTC(CRT_HWC_MODE,Inb_CRTC(CRT_HWC_MODE)&0xf2);break;
	}							/* SVS3 says 0 ... */

	Outb_CRTC(CRT_SYSCONTROL1,
		  (Inb_CRTC(CRT_SYSCONTROL1)&0xcf)
		  |CR50_32BPP);			/* bits per pixel*/

	Outb_CRTC(CRT_DACCONTROL,/* was 0x68*/(Inb_CRTC(CRT_DACCONTROL)&0x08)|0x40);


	h=(TM->x.width*1)/4+1; /* FIXME? - other bbps, s3drv says 8 ! */
	Outb_CRTC(CRT_MEMCONTROL4,CR61_L_ENABLE|(h>>8));
	Outb_CRTC(CRT_MEMCONTROL5,(h&0xff));


	Outb_CRTC(CRT_EXTMISC0,0x20);
	Outb_CRTC(CRT_EXTMISC1,0x10);

	return(0);
}  


/***************************************************************************/
/* *** ***        I n t e r f a c e   d e f i n i t i o n s        *** *** */
/***************************************************************************/

int kgi_chipset_check_mode(struct kgi_mode *TM,enum kgi_tm_cmd cmd)
{ 
	TM->realclockmul.mul=1;TM->realclockmul.div=1;
	TM->request.trflags|=TRF_NO_ACCEL;

	switch(TM->request.graphtype)
	{	case GT_TEXT		: 
			return(CheckText(TM,cmd));

		case GT_4BIT_PLANES	: 
			return(Check4bit(TM,cmd));
			
		case GT_8BIT		: 
			return(Check8bit(TM,cmd));
			
		case GT_15BIT		:
		case GT_16BIT		: 
			return(Check16bit(TM,cmd));

		case GT_24BIT		: 
			TM->request.graphtype=GT_32BIT;	/* Use 32 bpp instead */

		case GT_32BIT		:
			return(Check32bit(TM,cmd));

		default     		: 
			return(-1);
	}
}  

int kgi_chipset_set_mode(struct kgi_mode *TM)
{ 
	switch(TM->request.graphtype)
	{	case GT_TEXT		: 
			return( SetText(TM) );

		case GT_4BIT_PLANES	: 
			return( Set4bit(TM) );
			
		case GT_8BIT		: 
			return( Set8bit(TM) );
			
		case GT_15BIT		:
		case GT_16BIT		: 
			return( Set16bit(TM) );

		case GT_24BIT		: 
			TM->request.graphtype = GT_32BIT;	/* Use 32 bpp instead */

		case GT_32BIT		:
			return( Set32bit(TM) );

		default     		: 
			return(-1);
	}
}  


/****************************/
/*** Set start of display ***/
/****************************/

void hw_setdispstart(int dx,int dy)
{
	int x;

	if (CurrState->bgmode) return;
    
	x=dx+dy*CurrState->ModeNow.virtual.x;
  
	Outb_CRTC(0x0d,(x >>  2) & 0xff);    /*  sa2-sa9 */
	Outb_CRTC(0x0c,(x >> 10) & 0xff);    /* sa10-sa17 */
	Outb_CRTC(CRT_EXTCONTROL3,(x >> 18) & 0x0f);

	/* FIXME - why does this not work in graphics-mode ? 
	   Pixel-panning works fine in Text-Mode ... */
	if (CurrState->ModeNow.graphtype==GT_TEXT)
		switch(CurrState->ModeNow.textgrid.x)
		{ case  9:Outb_ATR_CTL(0x13,(Inb_ATR_CTL(0x13)&0xf0)|8);break;
		  default:Outb_ATR_CTL(0x13,(Inb_ATR_CTL(0x13)&0xf0)|0);break;
		  case  7:Outb_ATR_CTL(0x13,(Inb_ATR_CTL(0x13)&0xf0)|1);break;
		}
	else Outb_ATR_CTL(0x13,(Inb_ATR_CTL(0x13)&0xf0)|((x & 3)<<1));
}

int hw_pwrsave(void)
{
	if (!CurrState->bgmode)
	{	switch(*(int *)parameters)
		{ case PWR_OFF    :Outb_SEQ(0x0d,(Inb_SEQ(0x0d)&0x0f)|0x50);break;
		  case PWR_SUSPEND:Outb_SEQ(0x0d,(Inb_SEQ(0x0d)&0x0f)|0x40);break;
		  case PWR_STANDBY:Outb_SEQ(0x0d,(Inb_SEQ(0x0d)&0x0f)|0x10);break;
		  case PWR_ON	  :Outb_SEQ(0x0d,Inb_SEQ(0x0d)&0x0f);break;
                  default         :Outb_SEQ(0x0d,Inb_SEQ(0x0d)&0x0f);return -EINVAL;
                }
	}                                  
	return 0;
}

static inline void unmap_all(void);
int sel_page=0;

int setpage(void)
{	sel_page=*(int *)parameters;
	unmap_all();
	return(0);
}

int getpage(void)
{	
	int i;

	if ((i=verify_area(VERIFY_WRITE, *(void **)parameters,sizeof(int))))
		return(i);

	put_fs_long(sel_page,*(void **)parameters);
	return(0);
}
                                                 
int kgi_chipset_ioctl(struct inode *inode, struct file *file, \
        unsigned int cmd, unsigned long arg)
{ DEV_NUM_SETUP;

  switch(cmd) { 
        case IO_CHIP_SETVISWIN:
        	COPYPAR(2*sizeof(int));
    		return(setdispstart());
        case IO_CHIP_GETVISWIN:
        	COPYPAR(2*sizeof(int));
    		return(getdispstart());
        case IO_CHIP_SETPAGE:
        	COPYPAR(sizeof(int));
    		return(setpage());
        case IO_CHIP_GETPAGE:
        	COPYPAR(sizeof(int *));
    		return(getpage());
    	case IO_CHIP_WAITVRETRACE:
		while(inb(0x3da)&8);
		while((inb(0x3da)&8)==0);
		while(inb(0x3da)&8);
		while((inb(0x3da)&8)==0);
		return(0);

        default		       :return -ENODRVSUP_ALWAYS_LOWER;
   }
   /* We should never get here ... */
   return -ENODRVSUP_ALWAYS_CANT;
}                                                      

static int nopage_deadlock_check=0;

#define MAXMP 100

static struct Map { 
		struct task_struct *tsk;
		unsigned long lastmap; } map[MAXMP],*mp;

static int mapcnt=0;
static unsigned long mappedpage=-1;

static inline void unmap_all(void)
{ 	for(;mapcnt;mapcnt--)
	{	mp=&map[mapcnt-1];
		fast_unmap_page_range(mp->tsk,mp->lastmap, 0x10000);
	}
	mappedpage=-1;
}

static inline void page_select(int page)
{
	switch(CurrState->ModeNow.graphtype)
	{ case GT_4BIT_PLANES	: /* 4 Videopages per logical one */
		outw((page<<10)|CRT_EXTCONTROL4,0x3d4);break;/* Bank select */
	  case GT_8BIT		: /* Normal page flipping */
	  case GT_15BIT		: 
	  case GT_16BIT		: 
	  case GT_32BIT		:
	  	outw((page<<8)|CRT_EXTCONTROL4,0x3d4);break;
	  case GT_TEXT          : /* Can't page in Textmode */
	  default		:break; }
}

void kgi_linearmap_close(struct vm_area_struct * area)
{	
	nopage_deadlock_check=0;
	/* FIXME ... delete only those mappings that belong */
	/* to the closing task - not too critical, but ... */
	unmap_all();
}

unsigned long kgi_linearmap_nopage(struct vm_area_struct * area, unsigned long address,
	unsigned long page, int no_share)
{
	static pte_t *last_pte=NULL;
	static struct task_struct *tsk;
	
	VM_DEV_NUM_SETUP;
        	
	if (CurrState->bgmode) {	/* A backgrounded process accesses graphics mem ! */
		area->vm_task->blocked &= ~(1<<(SIGSEGV-1));
		send_sig(SIGSEGV, area->vm_task, 1);
        }

	if (nopage_deadlock_check && !pte_young(*last_pte) && tsk==area->vm_task) {
		area->vm_task->blocked &= ~(1<<(SIGBUS-1));
		send_sig(SIGBUS, area->vm_task, 1);
        }

        address -= area->vm_start;
	nopage_deadlock_check=!(address&0xffff);	/* fault at first byte of VGA page ? */
	address &= 0xffff0000;				/* which page to map */

	if (mappedpage!=address>>16)			/* delete all mapping for old page */
		unmap_all();

	/* VGA Granularity is 64 K  => address &= PAGE_MASK unnecessary */
	page_select(address>>16);
	address+= area->vm_start;

        map[mapcnt].tsk=area->vm_task;
        map[mapcnt].lastmap=address;
        if (mapcnt++>=MAXMP) {mapcnt--;printk("Whee - to many mappings - going mad !\n");}
        
	if (CurrState->ModeNow.graphtype==GT_TEXT)
		fast_remap_page_range(area->vm_task,address, 0xb8000, 0x10000, area->vm_page_prot);
	else
        	fast_remap_page_range(area->vm_task,address, 0xa0000, 0x10000, area->vm_page_prot);

        if (nopage_deadlock_check)
        	last_pte = pte_offset( pmd_offset( pgd_offset(tsk=area->vm_task, address), address), address);

	invalidate();

        ++area->vm_task->mm->min_flt;	/* do some statistics */
        
	return((unsigned long)NULL);
}

void kgi_pagedmap_close(struct vm_area_struct * area)
{	
	nopage_deadlock_check=0;
	/* FIXME ... delete only those mappings that belong */
	/* to the closing task - not too critical, but ... */
	unmap_all();
}


unsigned long kgi_pagedmap_nopage(struct vm_area_struct * area, unsigned long address,
	unsigned long page, int no_share)
{
	VM_DEV_NUM_SETUP;
        	
	if (CurrState->bgmode
	    || address -area->vm_start >= 64*1024) 
	{	/* A backgrounded process accesses graphics mem ! */
		area->vm_task->blocked &= ~(1<<(SIGSEGV-1));
		send_sig(SIGSEGV, area->vm_task, 1);
        }

	if (mappedpage!=sel_page)			/* delete all mapping for old page */
		unmap_all();

	/* VGA Granularity is 64 K  => address &= PAGE_MASK unnecessary */
	page_select(sel_page);

        map[mapcnt].tsk=area->vm_task;
        map[mapcnt].lastmap=address;
        if (mapcnt++>=MAXMP) {mapcnt--;printk("Whee - to many mappings - going mad !\n");}
        
	if (CurrState->ModeNow.graphtype==GT_TEXT)
		fast_remap_page_range(area->vm_task,address, 0xb8000, 0x10000, area->vm_page_prot);
	else
        	fast_remap_page_range(area->vm_task,address, 0xa0000, 0x10000, area->vm_page_prot);

	invalidate();

        ++area->vm_task->mm->min_flt;	/* do some statistics */
        
	return((unsigned long)NULL);
}

int kgi_chipset_init(void)
{
	outb(S3_DISABLE, S3_SUBSYSTEM);	/* chip wakeup */
	outb(S3_WAKEUP,  S3_SETUP);
	outb(S3_ENABLE,  S3_SUBSYSTEM);

	Detect_Vision964();		/* FIXME - should react to rc */
	
	unlock_s3();			/* Unlock s3-regs ... other Inits might need access */
  
	printk("Manufacturer : %s\n",chip_par.manufact);
	printk("Model        : %s\n",chip_par.model);
	printk("RAM / Maxsize: %d Bytes (%d x %d) \n",
        	chip_par.ramsize,
        	chip_par.maxdots.x,
        	chip_par.maxdots.y);

	_getsetfont_hw(OpenFont,0);
	return(0);			/* All went OK - install */
}

void kgi_chipset_done(void) 
{ 
	lock_s3();
}

