/*\
|*| ---------------------------------------------------------------------------
|*|	generic fixed clock chip driver
|*| ---------------------------------------------------------------------------
|*|
|*|	Copyright (C) 1995 by Steffen Seeger [seeger@physik.tu-chemnitz.de]
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	This program is distributed in the hope that it will be useful,
|*|	but WITHOUT ANY WARRANTY; without even the implied warranty of
|*|	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
|*|	GNU General Public License for more details.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*| ---------------------------------------------------------------------------
\*/

#if !(CLOCKVERSION == 01) 
 #warning	kgi_clock data structures changed, check this code !!
#endif

#define	FREQ	clock.clock.freq

/*\
|*|     GGI clock driver interface
\*/

int
kgi_clock_init (void)
{
  TRACE(printk("%s %s fixed clock driver.\n", clock.manufact, clock.model));
  return clock_init();
}

struct kgi_clock
*kgi_clock_get_info (void)
{ return &clock; }

int
kgi_clock_check_mode (struct kgi_mode *mode, enum kgi_tm_cmd cmd)
{
  int i = 0, err;

  while((FREQ[i] > mode->dclk) && (i < MAX_FIXED_CLOCKS))
    i++; 			/* search 1st value <= mode->dclk	*/

  if((i == MAX_FIXED_CLOCKS) || FREQ[i])
    i--;			/* out of bounds?			*/

  err = ((!i) && (mode->dclk > FREQ[0])) ?
	  ETM_DCLK_HIGH :
          (FREQ[i] > mode->dclk) ? 
	    ETM_DCLK_LOW : 
	    EOK;

  switch (cmd)
   {
     case TC_PROPOSE_TIMING:	/* set mode->dclk to next possible value     */
	mode->dclk  = FREQ[i]; mode->clock = -1;
	return (err == ETM_DCLK_HIGH) ? EOK : err;

     case TC_CHECK_TIMING:
	if (   (err == EOK) 	/* check if clock is near the discrete value */
	    && ((mode->dclk - FREQ[i]) < (mode->dclk*3/100)))
          {
	    mode->dclk = FREQ[i];	mode->clock = i;
	    return EOK;
          }
	 else
	  {
	    mode->clock = -1; 
	    return (err == EOK) ? ECLOCK : err;
	  }
     default:			/* this should not happen ... */
	INTERNALERROR;
	return ECLOCK;
   }
}

int
kgi_clock_set_mode (struct kgi_mode *mode)
{
  if ((mode->clock < 0) || (mode->dclk != FREQ[mode->clock]))
    {
      INTERNALERROR;
      return ECLOCK;
    }

  setCLK(index[mode->clock]);
  return EOK;
}

int
kgi_clock_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
	       unsigned long arg)
{
  return EINVAL;
}
