/*\
|*| --------------------------------------------------------------------------
|*|     kernel module interface definitions
|*| --------------------------------------------------------------------------
|*|
|*|	Copyright (C) 1995 by Steffen Seeger, [seeger@physik.tu-chemnitz.de] 
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	This program is distributed in the hope that it will be useful,
|*|	but WITHOUT ANY WARRANTY; without even the implied warranty of
|*|	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
|*|	GNU General Public License for more details.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*| ---------------------------------------------------------------------------
\*/
#ifndef _kgi_module_h
#define _kgi_module_h

#include "scrdrv.h"

/* Maximum number of consoles that can be under GGI-control 
   simultaneously - this will increase module size quite a bit */

#define MAX_CONS_NUM 8

#define DEV_NUM_SETUP CurrState=file->private_data

#define VM_DEV_NUM_SETUP CurrState=(struct ggi_StateInfo *)area->vm_offset

/*\
|*|	some useful debugging macros
\*/

#define INTERNALERROR printk("ggi: %s: %i: internal error, please report.\n",\
			      __FILE__,__LINE__);

#ifdef __KERNEL__
 #define NOT_TESTED \
    printk("\n\nGood !! You hit untested code in %s line %i. Please report whether\n"\
	   "there were any difficulties or not. (Simply send this message and a\n"\
	   "short explanation of your configuration to linux-ggi@eskimo.com.\n\n\n",\
	    __FILE__,__LINE__)
 #ifdef DEBUG
  #define TRACE(X) X
 #else
  #define TRACE(X)
 #endif
#endif

/*\
|*|	Interface definition
\*/

int	kgi_monitor_init (void);
void	kgi_monitor_done (void);
void	kgi_monitor_reset (void);
int	kgi_monitor_check_timing   (struct ggi_Timing *TM,int cmd);
int	kgi_monitor_set_timing     (struct ggi_Timing *TM);
int	kgi_monitor_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);


int	kgi_clock_init (void);
void	kgi_clock_done (void);
void	kgi_clock_reset (void);
int	kgi_clock_check_timing   (struct ggi_Timing *TM,int cmd);
int	kgi_clock_set_timing     (struct ggi_Timing *TM);
int	kgi_clock_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);


int	kgi_chipset_init (void);
void	kgi_chipset_done (void);
void	kgi_chipset_reset (void);
int	kgi_chipset_check_timing   (struct ggi_Timing *TM,int cmd);
int	kgi_chipset_set_timing     (struct ggi_Timing *TM);
int	kgi_chipset_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);


int	kgi_ramdac_init (void);
void	kgi_ramdac_done (void);
void	kgi_ramdac_reset (void);
int	kgi_ramdac_check_timing   (struct ggi_Timing *TM,int cmd);
int	kgi_ramdac_set_timing     (struct ggi_Timing *TM);
int	kgi_ramdac_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);


int	kgi_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);
int	kgi_mmap(struct inode * inode, struct file * file, struct vm_area_struct * vma);

/*\
|*|	generalized hardware IO functions.
\*/

void		outbDAC(unsigned char value, unsigned char dac_reg);
unsigned char	inbDAC(unsigned char dac_reg);
void		outsbDAC(unsigned short dac_reg, void *buf, int count);
void		insbDAC (unsigned short dac_reg, void *buf, int count);

void	setCLK(int value);
int	getCLK(void);

struct ggi_ChipParams ChipPar;

struct ggi_Coord    { int x,y  ; };

struct ggi_StateInfo 
	{ struct ggi_Timing_req	ModeNow;
	  int		      	bgmode,available,vt_num;
	  struct ggi_Drawinfo	*drawinfo;	/* Pointer to user-space drawinfo */
	  struct ggi_Coord	dispstart;
	  struct ggi_Palentry	palette[256];
	} StateNow[MAX_CONS_NUM];

struct ggi_StateInfo OpenState,*CurrState;

extern char parameters[256];	/* parameter transfer buffer */

#endif	/* _kgi_module_h */
