/*\
|*| --------------------------------------------------------------------------
|*|	kernel module interface definitions
|*| --------------------------------------------------------------------------
|*|
|*|	Copyright (C) 1995 by Steffen Seeger   [seeger@physik.tu-chemnitz.de]
|*|			  and Andreas Beck   [becka@hp.rz.uni-duesseldorf.de]
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	This program is distributed in the hope that it will be useful,
|*|	but WITHOUT ANY WARRANTY; without even the implied warranty of
|*|	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
|*|	GNU General Public License for more details.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*| ---------------------------------------------------------------------------
\*/

#ifndef _kgi_module_h
#define _kgi_module_h

#include "ggi.h"

/* Maximum number of consoles that can be under GGI-control 
   simultaneously - this will increase module size quite a bit */

#define MAX_CONS_NUM 8

#define DEV_NUM_SETUP CurrState=file->private_data

#define VM_DEV_NUM_SETUP CurrState=(struct kgi_StateInfo *)area->vm_offset

/*\
|*|	some useful debugging macros
\*/

#define INTERNALERROR printk("KGI: %s: %i: internal error, please report.\n",\
			      __FILE__,__LINE__);

#ifdef __KERNEL__
 #define NOT_TESTED \
    printk(\
    "\n\nGood !! You hit untested code in %s line %i. Please report whether\n"\
    "there were any difficulties or not. (Simply send this message and a\n"\
    "short explanation of your configuration to linux-ggi@eskimo.com.\n\n\n",\
    __FILE__,__LINE__)
 #ifdef DEBUG
  #define TRACE(X) X
 #else
  #define TRACE(X)
 #endif
#endif

/*\
|*|	generalized hardware IO functions.
\*/

	void	outbDAC(unsigned char value, unsigned char dac_reg);
unsigned char	inbDAC(unsigned char dac_reg);
	void	outsbDAC(unsigned short dac_reg, void *buf, int count);
	void	insbDAC (unsigned short dac_reg, void *buf, int count);

	void	setCLK(int value);	/* drive CS[0:3] lines to value	*/
	int	getCLK(void);		/* read state of CS[0:3] lines	*/

/*\
|*|	kernel driver data structures
\*/

#if (GGIVERSION != 01)
  #warning	common ggi structures changed, check kgi structures !!
#endif

#define	COMMONVERSION	01

	enum kgi_flags
	{	FL_MONITOR	= 0x000000FF,
			FL_MON_INTERLACE  = 0x00000001,	/* Ain't that rather chipset ? */
			FL_MON_POWERSAVE  = 0x00000002,
		FL_DAC		= 0x0000FF00,
			FL_DAC_8BIT	= 0x000001,	/* 8 or 6 bit dac */
			FL_DAC_2XCLOCK	= 0x000040,	/* internal 2*clk */
			FL_DAC_GAMMA	= 0x000080,	/* true color LUT */
		FL_CLOCK	= 0x00FF0000,
		FL_CHIPSET	= 0xFF000000,
			FL_CHIP_LMAP	   = 0x01000000,/* linear mapping */
			FL_CHIP_DOUBLESCAN = 0x02000000,
	};

	#define	KGI_COMMON char manufact[64],	/* manufacturer		*/\
				model[64];	/* model/trademark	*/\
	 		    int	modes;		/* supported modes	*/\
	         enum kgi_flags	flags;		/* special capabilities	*/\
	       struct ggi_coord	maxdots;	/* maximum resolution	*/

	struct	kgi_common
	{
	  KGI_COMMON
	};


#define	MONITORVERSION	01

	enum	kgi_display
	{
		EGA_8 = 0,	/* EGA monitor (8 colors)		*/
		EGA_16,		/* EGA monitor (16 colors)		*/
		VGA,		/* analog VGA monitor			*/
		PLASMA,		/* plasma display used on laptop's (???)*/
		TFT,		/* (???)				*/
		LCD,		/* Liquid Crystal Display		*/
		HERCMONO,	/* Monochrome, 3 cols (off,on,hilight)  */
		LAST_DISPTYPE,	/* last defined type			*/
					/* COLOR and MONO are flags	*/
		COLOR = 0,		/* able to display colors	*/
		MONO  = 0x80000000	/* b&w or grayscale display	*/
	};

	enum	kgi_sync
	{
		SYNC_NORMAL 	= 0,	/* normal syncing via separate lines  */
		SYNC_ON_RED     = 1, 	/* send sync on r,g,b line. These can */
		SYNC_ON_GREEN   = 2,	/* be ORed together.		      */
		SYNC_ON_BLUE    = 4,
		BLANK_PEDESTRAL = 8,
		POWER_SAVE	=16	/* has powersaveing via sync (VESA)   */
	};

	struct kgi_monitor 
	{
	   KGI_COMMON

	   struct ggi_coord size;	/* max. physical picture size in mm   */
	   enum kgi_display disptype;   /* display type			      */
	   enum kgi_sync    synctype;	/* synctype			      */
	   struct ggi_range pfreq;	/* pixel frequency range (Bandwidth)  */
	   struct ggi_range hfreq;	/* horizontal frequency range	      */
	   struct ggi_range vfreq;	/* vertical frequency range 	      */
	};


#define	CHIPSETVERSION	01

	struct kgi_chipset
	{
	  KGI_COMMON

	  int			ramsize;	/* installed RAM size	*/
	  struct ggi_range	clock;		/* DCLK range		*/
	};


#define RAMDACVERSION	01

	enum kgi_dac_type
	{
	    DAC_8BIT_PB   = 0x01,	/*   8 bit pixel bus		*/
	    DAC_16BIT_PB  = 0x02,	/*  16 bit pixel bus		*/
	    DAC_24BIT_PB  = 0x04,	/*  24 bit pixel bus		*/
	    DAC_32BIT_PB  = 0x08,	/*  32 bit pixel bus		*/
	    DAC_64BIT_PB  = 0x10,	/*  64 bit pixel bus		*/
	    DAC_128BIT_PB = 0x20,	/* 128 bit pixel bus		*/
	    DAC_PB_MASK   = 0x3F,	/* pixel bus width mask		*/
	    DAC_VRAM      = 0x80000000	/* VRAM shift clock support	*/
	};

	struct kgi_ramdac
	{
	  KGI_COMMON

	  enum   kgi_dac_type	dactype;	/* general device type	     */
	  struct ggi_range	normal;		/* DCLK range (normal mode)  */
	  struct ggi_range	dblclk;		/* DCLK (internal doubling)  */
          struct ggi_range	clock[GT_LAST];	/* DCLK ranges		     */
	};

	enum kgi_dacmode
	{ 
	  DAC_MUX_1	= 0,	/* DCLK range if  1:1 multiplex */
	  DAC_MUX_2	= 1,	/* DCLK range if  2:1 multiplex */ 
	  DAC_MUX_4	= 2,	/* DCLK range if  4:1 multiplex */
	  DAC_MUX_8	= 3,	/* DCLK range if  8:1 multiplex */
	  DAC_MUX_16   	= 4,	/* DCLK range if 16:1 multiplex */
	  DAC_MUX_32	= 5,	/* DCLK range if 32:1 multiplex */
	  DAC_MUX_DOUBLE= 8,	/* DCLK limit if internal clock doubling */
	  DAC_MUX_NORMAL= 9	/* DCLK limit if in normal mode */
	};


#define	CLOCKVERSION	01

	#define	MAX_FIXED_CLOCKS	16
	#define	MAX_PROG_CLOCKS		8

	enum kgi_clock_type
	{
	  CLK_TYPE_FIXED,	CLK_TYPE_PROG
	};

	struct kgi_clock
	{
	  KGI_COMMON

	  enum kgi_clock_type	type;
	  union
	  {
	     int		freq[MAX_FIXED_CLOCKS];
	     struct ggi_range	range[MAX_PROG_CLOCKS];		
	  } clock;
	};


#define	INFOVERSION	01

	union kgi_info		/* union to ensure proper mem allocation  */
	{
	  struct kgi_common	common;

	  struct kgi_monitor	monitor;
	  struct kgi_ramdac	ramdac;
	  struct kgi_clock	clock;
	  struct kgi_chipset	chipset;
	};


#define	MODEVERSION	01

	struct kgi_timing	/* timing parameters (in DCLK's or lines)  */
	{
	  unsigned short	width;		/* pixels/lines to display    */
	  unsigned short	blankstart;	/* assertion of BLANK	      */
	  unsigned short	syncstart;	/* assertion of SNYC	      */
	  unsigned short	syncend;	/* deassertion of SYNC	      */
	  unsigned short	blankend;	/* deassertion of BLANK	      */
	  unsigned short	total;		/* total pixels/lines 	      */
	  unsigned short	polarity;	/* sync polarity	      */
	};

	struct kgi_mode		/* mode setup info. (calculated or preset) */
	{
	  struct ggi_mode_request request;	/* requested mode	      */

	  /* These will be auto-calculated by the various drivers */
	  struct kgi_timing	x,y;		/* x and y timing parameter   */
	  unsigned short	ymagnify;	/* stretch ... times	      */
	  enum kgi_flags	flags;		/* special capabilities used  */
          struct ggi_ratio	realclockmul;   /* e.g. 9/8 for 9 bit textmode*/
	  int			dclk;		/* dotclock (master reference)*/
	  int			mclk;		/* memory clock		      */
          enum kgi_dacmode	muxmode;
        
	  int			monitor;	/* driver specific options    */
	  int			ramdac;
	  int			clock;
	  int			chipset;
	};

enum kgi_tm_rc {
		MR_OK=0,
		MR_LOWER_CLOCK,
		MR_LOWEST_CLOCK,
		MR_HIGHER_CLOCK,
		MR_HIGHEST_CLOCK,
		MR_X_TOO_SMALL,
		MR_Y_TOO_SMALL,
		MR_X_TOO_BIG,
		MR_Y_TOO_BIG };
		
enum kgi_tm_cmd {
		CMD_NONE=0,		/* Do what you like */
		CMD_CHECK,		/* Check only - do not change anything */
		CMD_MON_PERFECT,	/* get perfect timing */
		CMD_MON_SOME,		/* get any working timing */
		CMD_CLK_LOWER,		/* get lower clock */
		CMD_CLK_HIGHER		/* get higher clock */
		};

struct kgi_StateInfo 
	{ struct ggi_mode_request	ModeNow;
	  int		      	bgmode,available,vt_num;
	  struct ggi_drawinfo	*drawinfo;	/* Pointer to user-space drawinfo */
	  struct ggi_coord	dispstart;
	  struct ggi_color	palette[256];
	};


/*\
|*|	Hardware driver interfaces. Stick to these functions when interfacing
|*|	drivers !
\*/

int	kgi_monitor_init (void);
void	kgi_monitor_done (void);
void	kgi_monitor_reset (void);
struct	kgi_monitor *kgi_monitor_get_info (void);
int	kgi_monitor_check_mode(struct kgi_mode *mode, enum kgi_tm_cmd cmd);
int	kgi_monitor_set_mode (struct kgi_mode *mode);
int	kgi_monitor_ioctl (struct inode *inode, struct file *file,
			   unsigned int cmd, unsigned long arg);

int	kgi_clock_init (void);
void	kgi_clock_done (void);
void	kgi_clock_reset (void);
struct	kgi_clock *kgi_clock_get_info (void);
int	kgi_clock_check_mode (struct kgi_mode *mode, enum kgi_tm_cmd cmd);
int	kgi_clock_set_mode (struct kgi_mode *mode);
int	kgi_clock_ioctl (struct inode *inode, struct file *file,
			 unsigned int cmd, unsigned long arg);

int	kgi_chipset_init (void);
void	kgi_chipset_done (void);
void	kgi_chipset_reset (void);
struct	kgi_chipset *kgi_chipset_get_info (void);
int	kgi_chipset_check_mode (struct kgi_mode *mode, enum kgi_tm_cmd cmd);
int	kgi_chipset_set_mode (struct kgi_mode *mode);
int	kgi_chipset_ioctl (struct inode *inode, struct file *file,
			   unsigned int cmd, unsigned long arg);

int	kgi_ramdac_init (void);
void	kgi_ramdac_done (void);
void	kgi_ramdac_reset (void);
struct	kgi_ramdac *kgi_ramdac_get_info (void);
int	kgi_ramdac_check_mode (struct kgi_mode *mode, enum kgi_tm_cmd cmd);
int	kgi_ramdac_set_mode (struct kgi_mode *mode);
int	kgi_ramdac_ioctl (struct inode *inode, struct file *file,
			  unsigned int cmd, unsigned long arg);

int	kgi_ioctl(struct inode *inode, struct file *file,
		  unsigned int cmd, unsigned long arg);
int	kgi_mmap(struct inode * inode, struct file * file,
		 struct vm_area_struct * vma);

extern struct kgi_StateInfo StateNow[MAX_CONS_NUM];
extern struct kgi_StateInfo OpenState,*CurrState;
extern char parameters[256];	/* parameter transfer buffer */

#endif	/* _kgi_module_h */
