/*\
|*| --------------------------------------------------------------------------
|*|     GGI general definitons, structures, etc.
|*| --------------------------------------------------------------------------
|*|
|*|	Copyright (C) 1995 by Andreas Beck [becka@hp.rz.uni-duesseldorf.de]
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	This program is distributed in the hope that it will be useful,
|*|	but WITHOUT ANY WARRANTY; without even the implied warranty of
|*|	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
|*|	GNU General Public License for more details.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*| ---------------------------------------------------------------------------
\*/

#include <linux/fs.h>	/* Needed for some prototypes - so we include it */

/**
  *  Include file for General Graphics Interface
  **/

#define TRUE 1
#define FALSE 0

/* Generic substructures */
struct ggi_Range { int min,max; };
struct ggi_Ratio { int mul,div; };

/* Monitor definition stuff */

enum ggi_Disptype {	BW    = 0,
			EGA_8 = 8,
			EGA_16=16,
			VGA   =24,
			T_MONO= 1 };
/* The MONO suffix means grayscale monitor - or (|) it to the rest ... */

/* Some monitors require composite sync signals thus we should support 
   the video output wave form according to the RS-343A spec. */

enum ggi_Synctype {	SYNC_NORMAL     = 0,
			SYNC_ON_RED     = 1, 	/* these can be ORed together */
			SYNC_ON_GREEN   = 2,
			SYNC_ON_BLUE    = 4,
			BLANK_PEDESTRAL = 8,
			POWER_SAVE	=16};

struct ggi_MonParams {
	struct ggi_Range  pfreq;	/* Pixel Frequency (Bandwidth)*/
	struct ggi_Range  hfreq;	/* Horizontal Frequency range */
	struct ggi_Range  vfreq;	/* Vertical   Frequency range */
	int		  xmaxdots,
			  ymaxdots;	/* Maximum resolution (mask size) */
	int		  xsize,
			  ysize;	/* picture size in mm */
	char		  manufact[64],	/* manufacturer */
			  model[64];	/* model */
	enum ggi_Disptype disptype;	/* type of display */
        enum ggi_Synctype synctype;     /* type of syncing */
	};

/* Chipset specific stuff */


struct ggi_ChipParams {
	int			ramsize;	/* Installed (useable) RAM size */
	int			xmaxdots,	/* Maximum resolution */
				ymaxdots;	/* (register limits)  */
	struct ggi_Range	clock;          /* DCLC range */
	char			manufact[64],	/* manufacturer */
				model[64];	/* model */
	enum ggi_Graphtype	graphtypes;	/* supported modes (ored together) */
	};

/* Ramdac specific stuff */

enum ggi_DacFlags { DAC_8BIT    = 1,
                    DAC_2XCLOCK = 2,
                    DAC_GAMMA   = 4
                  };

enum ggi_DacType {
		    DAC_8BIT_PB = 0x01, /* 8 bit pixel bus, requires VGA mode */
		    DAC_16BIT_PB= 0x02, /* 16 bit pixel bus */
		    DAC_24BIT_PB= 0x04, /* 24 bit pixel bus */
		    DAC_32BIT_PB= 0x08, /* 32 bit pixel bus */
                    DAC_64BIT_PB= 0x10, /* 64 bit pixel bus */
		    DAC_128BIT_PB=0x20, /* 128 bit pixel bus */
		    DAC_PB_MASK = 0x3E, /* pixel bus width mask (not vga) */
		    DAC_VRAM    = 0x80000000 /* VRAM shift clock support */
		};

struct ggi_DacParams
 {
        char                    manufact[64]; /* manufacturer */
        char                    model[64];    /* model */
	enum   ggi_DacType	dactype;      /* general device type */
	enum   ggi_DacFlags     flags;        /* special device capabilities */
        struct ggi_Range        clock[10];    /* DCLK ranges      */
 };

enum ggi_DacRangeIndex { 
	DAC_MUX_1	= 0,	/* DCLK range if  1:1 multiplex */
	DAC_MUX_2	= 1,	/* DCLK range if  2:1 multiplex */ 
	DAC_MUX_4	= 2,	/* DCLK range if  4:1 multiplex */
	DAC_MUX_8	= 3,	/* DCLK range if  8:1 multiplex */
	DAC_MUX_16   	= 4,	/* DCLK range if 16:1 multiplex */
	DAC_MUX_32	= 5,	/* DCLK range if 32:1 multiplex */
	DAC_MUX_DOUBLE	= 8,	/* DCLK limit if internal clock doubling */
	DAC_MUX_NORMAL  = 9	/* DCLK limit if in normal mode */
};

struct ggi_Timing_req {
	/* This is requested by the user and
	   changed as necessary by the driver */
	int	xvisible,		/* visible window size */
		yvisible;
	int	xvirtual,		/* virtual resolution,  */
		yvirtual;		/* i.e. scrolling range */
	enum	ggi_Graphtype 	graphtype;
	int	xtextgrid,
		ytextgrid;		/* font size for text mode */
	enum ggi_trflags	trflags;
};

enum ggi_Flags {	FL_INTERLACE	=  1,
			FL_DOUBLESCAN	=  2 };	/* What more ? */

struct ggi_Timing {
	struct ggi_Timing_req tr;

	/* These will be auto-calculated by the various drivers */
	int	clock;			/* in Hz */
	struct	ggi_Ratio	realclockmul;	/* e.g. 9/8 for 9 bit textmode */
	int	xwidth,
		xsyncstart,
		xsyncend,
		xend;			/* like in XFree */
	int	ywidth,
		ysyncstart,
		ysyncend,
		yend;			/* like in XFree */
	int	xsyncpol,		/* X sync polarity */
		ysyncpol;		/* Y  "      "     */
	int 	ymagnify;		/* CRTC-9 bits 0-4  (+1) */
	enum	ggi_Flags	flags;	/* special things like INTERLACE,DOUBLE_SCAN */
	enum ggi_DacRangeIndex  muxmode;
	};

enum mr_rc {	MR_OK=0,
		MR_LOWER_CLOCK,
		MR_LOWEST_CLOCK,
		MR_HIGHER_CLOCK,
		MR_HIGHEST_CLOCK,
		MR_X_TOO_SMALL,
		MR_Y_TOO_SMALL,
		MR_X_TOO_BIG,
		MR_Y_TOO_BIG };
		
enum mr_cmd {	CMD_NONE=0,		/* Do what you like */
		CMD_MON_PERFECT,	/* get perfect timing */
		CMD_MON_SOME,		/* get any working timing */
		CMD_CLK_LOWER,		/* get lower clock */
		CMD_CLK_HIGHER		/* get higher clock */
		};

struct ggi_Palentry {unsigned char r,g,b;};

struct ggi_Drawinfo {	int fg_color;
			int bg_color;
			int drawmode; };

struct ggi_Cursorinfo {	int	col1,col2;
			int	sx,sy;
			int 	hotx,hoty;
			char	*color,*shape; };
