/* 
   Generic functionality. This does not depend on any particular hardware.

   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/tty.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/malloc.h>
#include <linux/fcntl.h>
#include <linux/vt.h>
#include <sys/ioctl.h>
#include <asm/io.h>
#include <asm/pgtable.h>
#include "graphdev.h"
#include "kgi-module.h"
#include "vga_generic.h"
#include "generic.h"
#include "generic-needs.h"

char OpenFont[8192];		/* We store the font of the textmode-console here */

struct ggi_StateInfo StateNow[MAX_CONS_NUM],*CurrState;

struct ggi_StateInfo OpenState
	= { {	80*8,25*16,
		1024,1024,
		GT_TEXT,
		8,16 },
		0,1,0,		/* Console is in FG-mode, available and "attached" to vt 0 */
	        NULL,		/* DrawInfo not present */
	    { 	0,0},{0,} };

char parameters[256];	/* parameter transfer buffer */


/*******************************************
*** Debugging routine for mode resolving ***
*******************************************/

#ifdef DEBUG
void dump_tm(struct ggi_Timing *tm,char *label,int x)
{ printk("%s : %x\n",label,x);
  printk("%6d: %5d %5d %5d %5d\n",
         tm->clock/1000,tm->xwidth,tm->xsyncstart,tm->xsyncend,tm->xend);
  printk("        %5d %5d %5d %5d\n",
         tm->ywidth,tm->ysyncstart,tm->ysyncend,tm->yend);
  printk("        %5dx%5d of %5dx%5d tg:%3dx%3d\n\n",
         tm->tr.xvisible ,tm->tr.yvisible ,
         tm->tr.xvirtual ,tm->tr.yvirtual ,
         tm->tr.xtextgrid,tm->tr.ytextgrid);
}
#endif

/******************************************/
/*** SET a mode - this will change soon ***/
/******************************************/
int _setmode(struct ggi_Timing *tm)
{ int rc;

  if ((rc=kgi_CheckChipsetTiming(tm,CMD_NONE))) return(rc);

  if (!CurrState->bgmode)
   { if (kgi_SetChipsetTiming  (tm)==-1) return -ENOMODESUP_CHIP; }
  return 0;
}

int setmode(void)
{ int i;struct ggi_Timing tm;

  if ((i=verify_area(VERIFY_READ , *(void **) parameters, sizeof(struct ggi_Timing_req)))||
      (i=verify_area(VERIFY_WRITE, *(void **) parameters, sizeof(struct ggi_Timing_req)))) 
         return(i);

  memcpy_fromfs(&tm.tr, *(void **) parameters, sizeof(struct ggi_Timing_req));

  if ((i=_setmode(&tm))) return i;

  memcpy_tofs  (*(void **) parameters,&tm.tr , sizeof(struct ggi_Timing_req));

  hw_setdispstart(0,0);
  CurrState->ModeNow=tm.tr;return(0);
}

int getmode(void)
{ int i;
  if ((i=verify_area(VERIFY_WRITE, *(void **) parameters, sizeof(struct ggi_Timing_req)))) 
         return(i);
  memcpy_tofs  (*(void **) parameters,&CurrState->ModeNow , sizeof(struct ggi_Timing_req));
  return(0);
}

int keepmode(void)
{ OpenState=*CurrState;return(0); }

int getdeffont(void)
{ int i;
  if ((i=verify_area(VERIFY_WRITE, *(void **) parameters, sizeof(OpenFont)))) 
         return(i);
  memcpy_tofs  (*(void **) parameters,OpenFont, sizeof(OpenFont));
  return(0);
}

int setdeffont(void)
{ int i;
  if ((i=verify_area(VERIFY_READ , *(void **) parameters, sizeof(OpenFont))))
         return(i);
  memcpy_fromfs(OpenFont, *(void **) parameters, sizeof(OpenFont));
  return(0);
}

int setdispstart(void)
{ CurrState->dispstart.x=*((int *)parameters);
  CurrState->dispstart.y=*((int *)parameters+1);
  hw_setdispstart(CurrState->dispstart.x,CurrState->dispstart.y);
  return 0; }

int getdispstart(void)
{ int i;
  if ((i=verify_area(VERIFY_WRITE, *(void **) parameters, sizeof(struct ggi_Coord)))) 
         return(i);
  memcpy_tofs  (*(void **) parameters,&CurrState->dispstart,sizeof(struct ggi_Coord));
  return(0);
}
                                                                                        
/* return size of state-record */
int getstatesize(void)
{ int i; 
  i = verify_area(VERIFY_WRITE, *(void **) parameters, sizeof(unsigned long));
  if (!i) put_fs_long(sizeof(struct ggi_StateInfo), *(unsigned long **)parameters);
  return i;
}

int savestate(void)
{ int i;
  if ((i=verify_area(VERIFY_WRITE, *(void **)parameters,sizeof(struct ggi_StateInfo)))) 
         return(i);
 memcpy_tofs(*(void **)parameters,CurrState,sizeof(struct ggi_StateInfo));
 return(0);
}

int _restorestate(struct ggi_StateInfo *State)
{ int i;
  struct ggi_Timing gt;

  gt.tr=State->ModeNow;
  if ((i=_setmode(&gt))) return(i);

  if ((i=hw_setpal (0,256,State->palette))) return(i);
  hw_setdispstart(State->dispstart.x,State->dispstart.y);

  if (State->ModeNow.graphtype==GT_TEXT) _getsetfont_hw(OpenFont,1);
 
  return(0);
}

int restorestate(void)
{ int i;
  if ((i=verify_area(VERIFY_READ, *(void **)parameters,sizeof(struct ggi_StateInfo)))) 
         return(i);
 memcpy_fromfs(CurrState,*(void **)parameters,sizeof(struct ggi_StateInfo));
 return(_restorestate(CurrState));
}

int bgmode(int onoff)
{ 
  CurrState->bgmode=0;			/* Really do it ! */
  if (onoff)
  { _restorestate(&OpenState);	/* Set state at open physically */
  } else
  { _restorestate(CurrState);	/* Reset to current state */
  }
  CurrState->bgmode=onoff;
  return(0);
}

int getvt(void)
{ int i;
  if ((i=verify_area(VERIFY_WRITE, *(void **)parameters,sizeof(int)))) 
         return(i);
 put_fs_long(CurrState->vt_num,*(void **)parameters);
 return(0);
}

/*****************************************************************************
****  Graphics  Device  Interface  Funtions                               ****
*****************************************************************************/

int kgi_Init(void)
{ int x;

  /* These have to fill in OpenState so we can revert to it */

  if ((x=kgi_ChipsetInit())) return(x);
  if ((x=kgi_ClockInit()  )) { kgi_ChipsetCleanup();return(x);}
  if ((x=kgi_RamdacInit() )) { kgi_ClockCleanup();
  			       kgi_ChipsetCleanup();return(x);}
  if ((x=kgi_MonitorInit())) { kgi_RamdacCleanup();
  			       kgi_ClockCleanup();
  			       kgi_ChipsetCleanup();return(x);}

  for(x=0;x<MAX_CONS_NUM;x++)
  { StateNow[x]=OpenState;}

  return(0);
}  				  

void kgi_Cleanup(void)
{ 
  kgi_MonitorCleanup();
  kgi_RamdacCleanup();
  kgi_ClockCleanup();
  kgi_ChipsetCleanup();
}  				  

/*****************************************************************************
*** Module loading stuff                                                   ***
*****************************************************************************/

int init_module(void)
{ int x;
  printk("Dumping driver V"VERSION" loading ...\n");
  kgi_Init();
  return -EIO;/* WE always fail ... */
}  				  

void cleanup_module(void) 
{}
