/*\
|*| ---------------------------------------------------------------------------
|*|	monitor driver for Sony CPD-1730 (and other modern multisync's ???)
|*| ---------------------------------------------------------------------------
|*|
|*|	Copyright (C) 1995 Steffen Seeger, [seeger@physik.tu-chemnitz.de]
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	This program is distributed in the hope that it will be useful,
|*|	but WITHOUT ANY WARRANTY; without even the implied warranty of
|*|	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
|*|	GNU General Public License for more details.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*| ---------------------------------------------------------------------------
\*/

#include <linux/kernel.h>

#include "graphdev.h"
#include "kgi.h"

#define MON_VERSION "0.11"

#ifdef DEBUG
static char *DISPTYPES[LAST_DISPTYPE]=
{ "EGA-8","EGA-16","VGA","PLASMA","TFT","LCD" };
#endif

#if !((MONITORVERSION == 01) && (MODEVERSION == 01))
  #warning kgi data structures changed, check this code !!
#endif

static struct kgi_monitor
monitor = {
		"Sony Corporation",	/* manufacturer		*/
		"CPD-1730",		/* model		*/
		GM_ALL,			/* supported modes	*/
		FL_MON_POWERSAVE |	/* special capabilities	*/
		FL_MON_INTERLACE, 

		{ 1024, 768 },		/* maximum resolution	*/

		{  382, 242 },		/* max. picture size	*/
		VGA | COLOR,		/* VGA color monitor	*/
		SYNC_NORMAL|POWER_SAVE,	/* normal syncing, VESA	*/
		{     0, 130000000 },	/* pixel clock range	*/
		{ 28000,     58000 },	/* hfreq range		*/
		{   55,        110 }	/* vfreq range		*/
	   };

/*\
|*|	GGI interface
\*/

int
kgi_monitor_init(void)
{
   TRACE(printk("%s %s Monitor Driver v"MON_VERSION"\n", \
         monitor.manufact, monitor.model));
   TRACE(printk("(%s - %s, %dx%d Pixel, %d mm x %d mm, %d-%d kHz, %d-%d Hz)\n",\
	  DISPTYPES[monitor.disptype & ~MONO],\
	  (monitor.disptype & MONO) ? "monochrome" : "color",\
	  monitor.maxdots.x,\
	  monitor.maxdots.y,\
	  monitor.size.x,\
	  monitor.size.y,\
	  monitor.hfreq.min/1000 ,monitor.hfreq.max/1000,\
	  monitor.vfreq.min      ,monitor.vfreq.max ));
   return TRUE;
}

void
kgi_monitor_done(void)
{
   TRACE(printk("%s %s Monitor driver removed.\n", \
         monitor.manufact, monitor.model));
}

struct kgi_monitor
*kgi_monitor_get_info (void)
{
  return &monitor;
}

int
kgi_monitor_check_mode (struct kgi_mode *mode, enum kgi_tm_cmd cmd)
{ 
  int hfreq, vfreq, result;
  register int width = mode->x.width;

  if (mode->dclk < monitor.pfreq.min)
    return ETM_DCLK_LOW;

  switch(cmd)
   {
     case TC_PROPOSE_TIMING:
	mode->dclk = monitor.pfreq.max;

	for(hfreq = monitor.hfreq.max; hfreq > monitor.hfreq.min;
	    mode->dclk -= 1000) /* Try from highest frequency down */
	  {
	    /* These timing calculations are designed to meet the VESA-specs
	       at 800 and 1024 pixels horizontal resolution. Anyhow, dclk
	       may not be near the specs. 
 	     */

	    mode->x.syncstart = width * (width      + 41360) /40000;
            mode->x.syncend   = width * (width*7/10 +  8952) / 8000;
            mode->x.total     = width * (width      +  9360) / 8000;
	    mode->x.blankstart= mode->x.syncstart;
	    mode->x.blankend  = mode->x.syncend;

            hfreq = mode->dclk / mode->x.total;
            if (monitor.hfreq.max < hfreq)
              continue;

            mode->y.syncstart = mode->y.width + 3;
            mode->y.syncend   = mode->y.width + 9;
            mode->y.total     = mode->y.width + hfreq/1000 + 1;/*1ms flyback*/
	    mode->y.blankstart= mode->y.syncstart;
	    mode->y.blankend  = mode->x.syncend;

            mode->x.polarity  = mode->y.polarity = 1; /* doesn't matter */

            vfreq = mode->dclk / (mode->x.total * mode->y.total);
            if ((monitor.vfreq.min <= vfreq) && (vfreq <= monitor.vfreq.max))
              {
		mode->request.size = monitor.size;
		TRACE(\
                 printk("%s: calculated %i Hz refresh (%i Hz, DCLK = %i Hz)\n",\
		         __FILE__, vfreq, hfreq, mode->dclk));
		return -EOK;
              }
          }
	TRACE(printk("%s: invalid mode request.\n",__FILE__));
	return ETM_DCLK_LOW;
	break;

     case TC_CHECK_TIMING: 
	hfreq = mode->dclk / mode->x.total;
	vfreq = mode->dclk / (mode->x.total * mode->y.total);
	
	result =   (monitor.hfreq.min <= hfreq) && (hfreq <= monitor.hfreq.max) 
	        && (monitor.vfreq.min <= vfreq) && (vfreq <= monitor.vfreq.max)
		&& (monitor.pfreq.min <= mode->dclk) 
		&& (mode->dclk <= monitor.pfreq.max);

	#ifdef DEBUG
	  if (result)
 	    TRACE(printk("%s: accepted %i Hz refresh (%i Hz, DCLK = %i Hz)\n",\
		__FILE__, vfreq, hfreq, mode->dclk));
	   else
	    TRACE(printk("%s: refresh check failed. (%i Hz, %i Hz, %i Hz)\n",\
		  __FILE__, vfreq, hfreq, mode->dclk));
	#endif

	return result ? EOK : EMONITOR;
	break;

     default:
	INTERNALERROR;
	return  EMONITOR;
   }
}

int
kgi_monitor_set_mode(struct kgi_mode *TM)
{ 
   return EOK;
}

int
kgi_monitor_ioctl (struct inode *inode, struct file *file, unsigned int cmd,
		  unsigned long arg)
{
  return EMONITOR;
}
