/* 
   Generic Monitor driver for modern Multi-Sync-type monitors.

   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
 
#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/malloc.h>
#include <linux/fcntl.h>
#include <linux/vt.h>
#include <sys/ioctl.h>
#include <asm/io.h>
#include <asm/pgtable.h>
#include "graphdev.h"
#include "kgi.h"
#include "multisync_config.h"
#include "monitor/monitor_generic.h"

static char *DISPTYPES[4]={ "B&W","EGA-8","EGA-16","VGA" };

int  kgi_monitor_check_mode  (struct kgi_mode * MT,enum kgi_tm_cmd cmd)
{ int flg,acfreq,h;
  int err=0;

  /* We ignore CMD right now and always want to be perfect ... FIXME */
  
  /* Lower clock if out of bounds ... */
  for(flg=0;MT->clock>MT->x.width*mon_par.hfreq.min;MT->clock-=1000,flg++)

  { h=MT->clock/1000;			/* Pixels for 1 ms */

    MT->x.total      =(MT->x.width+
    		    (h*MonTimeConst.xafter_ns)/1000000+
    		    MonTimeConst.xafter_pix)
    		   &0xfffffff8;
    while(MT->x.total<=MT->x.width+16) MT->x.total+=8;

    acfreq=MT->clock/MT->x.total;			/* Lines for 1 us */

    if (acfreq>mon_par.hfreq.max) 
      if (err==MR_LOWEST_CLOCK) return(MR_X_TOO_SMALL);
      else continue;				/* Try lower clock */

    if (acfreq<mon_par.hfreq.min) 
      if (flg==0) return(MR_X_TOO_BIG);		/* Screen too wide even at highest clock */
      else        return(MR_Y_TOO_SMALL);	/* aspect ratio is bad ... */
      
    /* Here to increase performance - these do not matter for line frequency */
    MT->x.syncstart=(MT->x.width+
    		    (h*MonTimeConst.xbefore_ns)/1000000+
    		    MonTimeConst.xbefore_pix)
    		   &0xfffffff8;		/* must be a multiple of 8 */
    while(MT->x.syncstart<=MT->x.width) MT->x.syncstart+=8;

    MT->x.syncend  =(MT->x.width+
                    (h*MonTimeConst.xlen_ns)/1000000+
                    MonTimeConst.xlen_pix)
                   &0xfffffff8;
    while(MT->x.syncend<=MT->x.syncstart) MT->x.syncend+=8;

    h=acfreq;					/* Lines for 1 s */
    
    MT->y.total      =MT->y.width+MonTimeConst.yafter_pix
    			     +(MonTimeConst.yafter_us*h)/1000000; /*flyback */
    while(MT->y.total<=MT->y.width+2) MT->y.total++;

    acfreq=MT->clock/(MT->x.total*MT->y.total);
    
    if (acfreq>mon_par.vfreq.max)
      if (err==MR_LOWEST_CLOCK) return(MR_Y_TOO_SMALL);
      else continue;				/* Try lower clock */

    if (acfreq<mon_par.vfreq.min) return(MR_Y_TOO_BIG);		/* Screen too tall */

    MT->y.syncstart=MT->y.width+MonTimeConst.ybefore_pix
    			     +(MonTimeConst.ybefore_us*h)/1000000;
    while(MT->y.syncstart<=MT->y.width) MT->y.syncstart++;
    MT->y.syncend  =MT->y.width+MonTimeConst.ylen_pix
    			     +(MonTimeConst.ylen_us*h)/1000000;
    while(MT->y.syncend<=MT->y.syncstart) MT->y.syncend++;

    MT->x.polarity  =MT->y.polarity		= 1;	/* No matter for this monitor */

    switch(err=kgi_clock_check_mode(MT,CMD_CLK_LOWER))
    { case MR_LOWEST_CLOCK:
      case MR_LOWER_CLOCK :break;
      default: return err;
    }
  }
  return(-ENOMODESUP_MON);
}

int  kgi_monitor_set_mode(struct kgi_mode * MT) {}


int kgi_monitor_ioctl(struct inode *inode, struct file *file, \
        unsigned int cmd, unsigned long arg)
{ switch(cmd)
  { case IO_MONITOR_GETINFO    :COPYPAR(sizeof(void *));
			     return(mon_getinfo());
    case IO_MONITOR_SETINFO    :if (!suser()) return(-EPERM);
			     COPYPAR(sizeof(void *));
			     return(mon_setinfo());

    case IO_MONITOR_SETPREFMODE:if (!suser()) return(-EPERM);
			     COPYPAR(sizeof(void *));
			     return(mon_setprefmode());
    case IO_MONITOR_GETPREFMODE:COPYPAR(sizeof(int)+sizeof(void *));
			     return(mon_getprefmode());
    case IO_MONITOR_CLEARPREF  :if (!suser()) return(-EPERM);
                             TMnum=0;return(0);

    case IO_MONITOR_POWERSAVE  :/* if (!suser()) return(-EPERM);  Should this be suser only ??? */
    			     COPYPAR(sizeof(int));
    			     return(hw_pwrsave());
    default : return -ENODRVSUP_ALWAYS_CANT; }
  return -ENODRVSUP_ALWAYS_CANT;
}
          

/* Argument parsing and initialization */

int kgi_monitor_init(void)
{ printk("Monitor driver V"VERSION" loading ...\n");
  printk("Manufacturer : %s\n",mon_par.manufact);
  printk("Model        : %s\n",mon_par.model);
  printk("Type         : %s - %s  [%dx%d]\n",
  	  DISPTYPES[mon_par.disptype/8],
  	  (mon_par.disptype&1) ? "Monochrome" : "Color",
  	  mon_par.maxdots.x,
  	  mon_par.maxdots.y);
  printk("Size         : %d mm x %d mm\n",
  	  mon_par.size.x,
  	  mon_par.size.y);
  printk("Frequencies  : %d-%d kHz , %d-%d Hz\n",
  	  mon_par.hfreq.min/1000 ,mon_par.hfreq.max/1000 ,
  	  mon_par.vfreq.min      ,mon_par.vfreq.max );
  return(0);	/* No chance to check if this is right ... */
}  				  

void kgi_monitor_done(void) {}

void kgi_monitor_reset(void) { /* How do you reset a monitor ? */ }