/*\
|*| ---------------------------------------------------------------------------
|*|	ATT20c505 / ATT20c504 RAMDAC driver
|*| ---------------------------------------------------------------------------
|*|
|*|	Copyright (C) 1995 by Steffen Seeger, [seeger@physik.tu-chemnitz.de]
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	This program is distributed in the hope that it will be useful,
|*|	but WITHOUT ANY WARRANTY; without even the implied warranty of
|*|	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
|*|	GNU General Public License for more details.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*| ---------------------------------------------------------------------------
\*/

#include <linux/kernel.h>

#include "graphdev.h"
#include "kgi.h"
#include "ATT/20c505.h"

#include "default/gamma.inc"
#include "default/vga.inc"
#include "VGA/clut.inc"

#define DAC_VERSION "0.10"
#define PCLK DAC_C2_PCLK0     /* or DAC_C2_PCLK1 ? (doesn't matter on 20SV) */

#if !((MONITORVERSION == 01) && (MODEVERSION == 01) && (RAMDACVERSION == 01))
  #warning kgi datastructures have changed, check this code !!
#endif

enum RAMDAC_types 
 {
	ATT20C504_85	= 0,
	ATT20C504_11	= 1,

	ATT20C505_11	= 2,
	ATT20C505_13	= 3,
	ATT20C505_17	= 4,

	DAC_UNKNOWN	= -1
 };

#define ATT20C504 ATT20C504_85
#define ATT20C505 ATT20C505_11

#define	SUPPORTED (GM_TEXT | GM_1BIT | GM_4BIT | GM_8BIT \
		  | GM_15BIT | GM_16BIT | GM_32BIT)

static struct kgi_ramdac att_params[5] =
      {
       { "AT&T Microelectronics",	/* manufacturer		*/
	 "ATT 20C504-85",		/* model		*/
	 SUPPORTED,			/* supported modes	*/
	 FL_DAC_8BIT | FL_DAC_GAMMA,	/* flags		*/
	 { 0x0FFF, 0x0FFF },		/* maximum resolution	*/

	 DAC_8BIT_PB | 			/* device type		*/
	 DAC_32BIT_PB | DAC_VRAM,
	  {       0,  85000000 },	/* normal mode		*/
	  {       0,         0 },	/* 2xclk mode		*/

	 {{	  0,  75000000 },	/* 	GT_TEXT		*/
	  {       0,  85000000 },	/* 	GT_1BIT		*/
	  {	  0,  85000000 },	/* 	GT_4BIT		*/
	  {       0,  85000000 },	/* 	GT_8BIT		*/	
	  {	  0,  85000000 },	/* 	GT_15BIT	*/
	  { 	  0,  85000000 },	/* 	GT_16BIT	*/
	  {       0,         0 },	/* 	GT_24BIT	*/
	  {	  0,  75000000 }}	/* 	GT_32BIT	*/
       },

       { "AT&T Microelectronics",	/* manufacturer		*/
	 "ATT 20C504-11",		/* model		*/
	 SUPPORTED,			/* supported modes	*/
	 FL_DAC_8BIT | FL_DAC_GAMMA,	/* flags		*/
	 { 0x0FFF, 0x0FFF },		/* maximum resolution	*/

	 DAC_8BIT_PB | 			/* device type		*/
	 DAC_32BIT_PB | DAC_VRAM,
	  {       0, 110000000 },	/* limit in normal mode	*/
	  {       0,         0 },	/* limit in 2xclk mode	*/
 
	 {{	  0,  75000000 },	/* 	GT_TEXT		*/
	  {       0, 110000000 },	/* 	GT_1BIT		*/
	  {	  0, 110000000 },	/* 	GT_4BIT		*/
	  {       0, 110000000 },	/* 	GT_8BIT		*/	
	  {	  0, 110000000 },	/* 	GT_15BIT	*/
	  { 	  0, 110000000 },	/* 	GT_16BIT	*/
	  {       0,         0 },	/* 	GT_24BIT	*/
	  {	  0,  75000000 }}	/* 	GT_32BIT	*/
       },

       { "AT&T Microelectronics",	/* manufacturer		*/
	 "ATT 20C505-11",		/* model		*/
	 SUPPORTED,			/* supported modes	*/
	 FL_DAC_8BIT | FL_DAC_GAMMA |	/* flags		*/
	 FL_DAC_2XCLOCK,
	 { 0x0FFF, 0x0FFF },		/* maximum resolution	*/

	 DAC_8BIT_PB | 			/* device type		*/
	 DAC_32BIT_PB | DAC_VRAM,
	  {       0, 110000000 },	/* normal mode		*/
	  {45000000,  55000000 },	/* limit in normal mode	*/

	 {{	  0,  75000000 },	/* 	GT_TEXT		*/
	  {       0, 110000000 },	/* 	GT_1BIT		*/
	  {	  0, 110000000 },	/* 	GT_4BIT		*/
	  {       0, 110000000 },	/* 	GT_8BIT		*/	
	  {	  0, 110000000 },	/* 	GT_15BIT	*/
	  { 	  0, 110000000 },	/* 	GT_16BIT	*/
	  {       0,         0 },	/* 	GT_24BIT	*/
	  {	  0,  75000000 }}	/* 	GT_32BIT	*/
       },

       { "AT&T Microelectronics",	/* manufacturer		*/
	 "ATT 20C505-13",		/* model		*/
	 SUPPORTED,			/* supported modes	*/
	 FL_DAC_8BIT | FL_DAC_GAMMA | 	/* flags		*/
	 FL_DAC_2XCLOCK,
	 { 0x0FFF, 0x0FFF },		/* maximum resolution	*/

	 DAC_8BIT_PB | 			/* device type		*/
	 DAC_32BIT_PB | DAC_VRAM,
	  {       0, 110000000 },	/* normal mode		*/
	  {45000000,  67500000 },	/* 2xclk mode		*/

	 {{	  0,  85000000 },	/* 	GT_TEXT		*/
	  {       0, 135000000 },	/* 	GT_1BIT		*/
	  {	  0, 135000000 },	/* 	GT_4BIT		*/
	  {       0, 135000000 },	/* 	GT_8BIT		*/	
	  {	  0, 135000000 },	/* 	GT_15BIT	*/
	  { 	  0, 135000000 },	/* 	GT_16BIT	*/
	  {       0,         0 },	/* 	GT_24BIT	*/
	  {	  0,  85000000 }}	/* 	GT_32BIT	*/
       },

       { "AT&T Microelectronics",	/* manufacturer		*/
	 "ATT 20C505-17",		/* model		*/
	 SUPPORTED,			/* supported modes	*/
	 FL_DAC_8BIT | FL_DAC_GAMMA |	/* flags		*/
	 FL_DAC_2XCLOCK,
	 { 0x0FFF, 0x0FFF },		/* maximum resolution	*/

	 DAC_8BIT_PB | 			/* device type		*/
	 DAC_32BIT_PB | DAC_VRAM,
	  { 	  0, 110000000 },	/* normal mode		*/
	  {45000000,  85000000 },	/* 2xclk mode		*/

	 {{	  0,  85000000 },	/* 	GT_TEXT		*/
	  {       0, 135000000 },	/* 	GT_1BIT		*/
	  {	  0, 175000000 },	/* 	GT_4BIT		*/
	  {       0, 175000000 },	/* 	GT_8BIT		*/	
	  {	  0, 175000000 },	/* 	GT_15BIT	*/
	  { 	  0, 175000000 },	/* 	GT_16BIT	*/
	  {       0,         0 },	/* 	GT_24BIT	*/
	  {	  0,  85000000 }}	/* 	GT_32BIT	*/
       }
      };

static struct kgi_ramdac *dac_params = NULL;
static unsigned char CR0,CR1,CR2,CR3;
static struct 
{
  unsigned char CR0,CR1,CR2,CR3;
  struct ggi_color clut[256];
} save;

/*\
|*|	GGI interface
\*/

int
kgi_ramdac_init(void)
{
   unsigned char status = inbDAC(DAC_STATUSREG);

   TRACE(printk("ATT RAMDAC driver v"DAC_VERSION"\n"));

   get_CLUT_entries(save.clut, 0, 256);
   save.CR0 = inbDAC(DAC_CONTROL0);
   save.CR1 = inbDAC(DAC_CONTROL1);
   save.CR2 = inbDAC(DAC_CONTROL2);
   outbDAC(save.CR0 | DAC_C0_EXT_UNLOCK, DAC_CONTROL0);
   outbDAC(DAC_PW_CR3_INDEX, DAC_PW_INDEX);
   save.CR3 = inbDAC(DAC_STATUSREG);   
   outbDAC(save.CR0, DAC_CONTROL0);

   if(status & DAC_ST_ATT_DEVICE)
     switch(status & DAC_ST_ID_MASK)
      {
	case DAC_ATT20C504:
		dac_params = att_params + ATT20C504;
		break;

	case DAC_ATT20C505:
		dac_params = att_params + ATT20C505;
		break;

	case DAC_ATT20C506:
		printk("%s: don't know how to handle ATT20C506 device.\n",\
			__FILE__);
		return FALSE;
	default:
		printk("%s: unknown chip ID (%.2x).\n",__FILE__,\
			 status & DAC_ST_ID_MASK);
		return FALSE;
      }
    else
     {
       printk("%s: no ATT RAMDAC detected.\n",__FILE__);
       return FALSE;
     }

  TRACE(printk("  %s %s detected.\n", dac_params->manufact, dac_params->model));
  return TRUE;
}

void
kgi_ramdac_done(void)
{
  outbDAC(CR0 | DAC_C0_EXT_UNLOCK | DAC_C0_SLEEP_MODE, DAC_CONTROL0);
  outbDAC(DAC_PW_CR3_INDEX, DAC_PW_INDEX);
  outbDAC(save.CR3, DAC_STATUSREG);
  outbDAC(save.CR2, DAC_CONTROL2);
  outbDAC(save.CR1, DAC_CONTROL1);
  outbDAC(save.CR0, DAC_CONTROL0);

  set_CLUT_entries(save.clut, 0, 256);

  TRACE(printk("%s %s driver removed.\n", dac_params->manufact,\
        dac_params->model));
}

struct kgi_ramdac
*kgi_ramdac_get_info (void)
{
  return dac_params; 
}

int
kgi_ramdac_check_mode (struct kgi_mode *mode, enum kgi_tm_cmd cmd)
{
  int div = -1;
  int newclock = mode->dclk;

  switch(mode->request.graphtype)
   {
     case GT_TEXT:	div = 0; break;
     case GT_1BIT:	div = 5; break;
     case GT_4BIT:	div = 3; break;
     case GT_8BIT:	div = 2; break;
  
     case GT_15BIT:
     case GT_16BIT:	div = 1; break;
  
     case GT_32BIT:	div = 0; break;
  
     case GT_24BIT:	return ERAMDAC;
     default:
  	TRACE(printk("%s: could not handle graphtype %i\n",\
  	     __FILE__, mode->request.graphtype));
 	return ERAMDAC;
   }

  switch (cmd)
   {
     case TC_PROPOSE_TIMING:
	if (newclock < dac_params->clock[mode->request.graphtype].min)
	  {
	    TRACE(printk("%s: invalid clock (%i).\n", __FILE__, newclock));
	    return ETM_DCLK_LOW;
	  }

	if (dac_params->clock[mode->request.graphtype].max < newclock)
    	  newclock = dac_params->clock[mode->request.graphtype].max;

	if (dac_params->normal.max < newclock)
	  {
	    TRACE(printk("%s: clock doubling mode not implemented yet.\n",\
			 __FILE__));
	    newclock = dac_params->normal.max;
	  }

	mode->dclk   = newclock;
	mode->clkdiv = div;
  	TRACE(printk("%s: %i:1 multiplex mode, DCLK = %i Hz\n", __FILE__,\
			    1 << mode->clkdiv, mode->dclk));

	mode->flags &= ~FL_DAC_2XCLOCK;
	mode->flags |= FL_DAC_8BIT;
	if((mode->request.graphtype == GT_32BIT) || 
	   (mode->request.graphtype == GT_15BIT) ||
	   (mode->request.graphtype == GT_16BIT))
	 mode->flags |= FL_DAC_GAMMA;
	
  	return EOK;

     case TC_CHECK_TIMING:
	if (mode->clkdiv != div)
	  return ERAMDAC;

	if (   (mode->dclk < dac_params->normal.min)
	    || (mode->dclk < dac_params->clock[mode->request.graphtype].min))
	  return ETM_DCLK_LOW;

	if (   (mode->dclk > dac_params->normal.max)
	    || (mode->dclk > dac_params->clock[mode->request.graphtype].max))
	  return ETM_DCLK_HIGH;

	return EOK;

    default:
	INTERNALERROR;
	return ERAMDAC;
  }
}

int
kgi_ramdac_set_mode(struct kgi_mode *mode)
{
  void *clut;
  struct kgi_monitor *monitor;

  /*\
  |*|	setup CR0
  \*/
     CR0 = DAC_C0_8BIT;		/* we'll always use 8 bit DAC */
     monitor = kgi_monitor_get_info();
     if (monitor->synctype & SYNC_ON_RED)     CR0 |= DAC_C0_SYNC_ON_RED;
     if (monitor->synctype & SYNC_ON_GREEN)   CR0 |= DAC_C0_SYNC_ON_GREEN;
     if (monitor->synctype & SYNC_ON_BLUE)    CR0 |= DAC_C0_SYNC_ON_BLUE;
     if (monitor->synctype & BLANK_PEDESTRAL) CR0 |= DAC_C0_BLANK_LEVEL;


  /*\
  |*|	setup CR1
  \*/
     switch(mode->request.graphtype)
      {
        case GT_TEXT:	CR1 = 0;		clut = VGAPalette; break;
        case GT_1BIT:	CR1 = DAC_C1_1BPP;	clut = VGAPalette; break;
	case GT_4BIT:	CR1 = DAC_C1_4BPP;	clut = VGAPalette; break;
	case GT_8BIT:	CR1 = DAC_C1_8BPP;	clut = VGAPalette; break;
	case GT_15BIT:	CR1 = DAC_C1_15BPP;	clut = GammaMap;   break;
	case GT_16BIT:	CR1 = DAC_C1_16BPP;	clut = GammaMap;   break;
	case GT_32BIT:	CR1 = DAC_C1_24BPP;	clut = GammaMap;   break;

	case GT_24BIT:
	default:	INTERNALERROR; return ERAMDAC;
      };

     CR1 |= DAC_C1_CLUT_BYPASS;

  /*\
  |*|	setup CR2
  \*/
      CR2 =  ((mode->request.graphtype == GT_TEXT) ? 0:  DAC_C2_ENHANCED_MODE)
           | ((mode->flags & FL_MON_INTERLACE)    ? DAC_C2_INTERLACED : 0)
           | PCLK;

  /*\
  |*|	setup CR3
  \*/
     CR3 = 0;		/* we have no clock doubling yet ... */

  /*\
  |*|	program CR3,CR2,CR1,CR0 (CR3 looks a bit strange, but is correct ;-)
  \*/
     outbDAC(CR0 | DAC_C0_EXT_UNLOCK | DAC_C0_SLEEP_MODE, DAC_CONTROL0);
     outbDAC(DAC_PW_CR3_INDEX, DAC_PW_INDEX);
     outbDAC(CR3, DAC_STATUSREG);

     outbDAC(CR2, DAC_CONTROL2);
     outbDAC(CR1, DAC_CONTROL1);
     outbDAC(CR0, DAC_CONTROL0);

  mode->flags |= (clut == GammaMap) ? FL_DAC_GAMMA : 0;
  mode->flags |= FL_DAC_8BIT;
    
  set_CLUT_entries(clut, 0, 256);	/* set standard clut */
  return TRUE;
}


int
kgi_ramdac_ioctl (struct inode *inode, struct file *file, unsigned int cmd,
		  unsigned long arg)
{
  /* this may set
       -  dac_params (take care if 504 or 505 was detected).
       -  hardware cursor
       -  CLUT (or gamma correction table).
   */
  return -1;
}

