/*\
|*| ---------------------------------------------------------------------------
|*|	BrookTree BT485 RAMDAC driver
|*| ---------------------------------------------------------------------------
|*|
|*|	Copyright (C) 1995 by Andreas Beck,[becka@sunserver1.rz.uni-duesseldorf.de]
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	This program is distributed in the hope that it will be useful,
|*|	but WITHOUT ANY WARRANTY; without even the implied warranty of
|*|	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
|*|	GNU General Public License for more details.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*| ---------------------------------------------------------------------------
\*/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/malloc.h>
#include <linux/fcntl.h>
#include <linux/vt.h>
#include <sys/ioctl.h>
#include <asm/io.h>
#include <asm/pgtable.h>
#include "graphdev.h"
#include "kgi.h"
#include "bt485.h"
#include "default/vga.inc"
#include "default/gamma.inc"

static struct kgi_ramdac ramdac_info =
       { "BrookTree", 			/* Manufacturer */
         "BT 485",			/* Model */
         0,				/* Supported modes - depr. */
	 FL_DAC_8BIT			/* We have 8 bit mode */
         | FL_DAC_2XCLOCK		/* have clock doubler */
         /* | FL_DAC_GAMMA*/,		/* Have gamma correction ? */
	 {0,0},				/* DAC doesn't limit resolution */
	 DAC_8BIT_PB | DAC_32BIT_PB | DAC_VRAM,

         {       0,  135000000},
	 {       0,          0},/* FIXME ! */

         {{       0,   75000000},
	  {       0,  135000000},
          {       0,  135000000},
          {       0,  135000000},
	  {       0,  135000000},
          {       0,  135000000},
	  {       0,  135000000} }
       };
       
struct ggi_cursorinfo ci = { 	0x0000ff,0x00ff00,
				64,64,
				0,0,
				NULL,NULL };

/**************************/
/*** Palette operations ***/
/**************************/
int hw_setpal(int start,int anz,struct ggi_color *buf)
{ if (CurrState->bgmode) return(0);
  outbDAC(start,BT_WRITE_ADDR);
  while(anz--)
  { outbDAC(buf->r,BT_RAMDAC_DATA);
    outbDAC(buf->g,BT_RAMDAC_DATA);
    outbDAC(buf->b,BT_RAMDAC_DATA);
    buf++; }
 return(0);
}

int setpal(void)
{ int i;
  struct MyP {	int start,anz;
		struct ggi_color *pal; } *mypars;

  mypars=(struct MyP *)parameters;

  if ((i=verify_area(VERIFY_READ, mypars->pal, 
                     sizeof(struct ggi_color)*mypars->anz))) 
         return(i);

  if (mypars->anz  <  0||
      mypars->start<  0||
      mypars->start>255||
      mypars->anz+mypars->start>256) return(-EINVAL);

  memcpy_fromfs(&CurrState->palette[mypars->start],mypars->pal,
  		sizeof(struct ggi_color)*mypars->anz);

  hw_setpal(mypars->start,mypars->anz,&CurrState->palette[mypars->start]);
  return(0);
}

void hw_getpal(int start,int anz,struct ggi_color *buffer)
{ outbDAC(start,BT_READ_ADDR);
  while(anz--)
  { buffer->r=inbDAC(BT_RAMDAC_DATA);
    buffer->g=inbDAC(BT_RAMDAC_DATA);
    buffer->b=inbDAC(BT_RAMDAC_DATA);
    buffer++; }
}

int getpal(void)
{ int i;
  struct MyP {	int start,anz;
		struct ggi_color *pal; } *mypars;

  mypars=(struct MyP *)parameters;

  if ((i=verify_area(VERIFY_WRITE, mypars->pal, 
                     sizeof(struct ggi_color)*mypars->anz))) 
         return(i);

  if (mypars->anz  <  0||
      mypars->start<  0||
      mypars->start>255||
      mypars->anz+mypars->start>256) return(-EINVAL);

 memcpy_tofs(mypars->pal,&CurrState->palette[mypars->start], sizeof(struct ggi_color)*mypars->anz);
 return(0);
}

int
kgi_ramdac_check_mode(struct kgi_mode *TM,enum kgi_tm_cmd cmd)
{
  enum kgi_dacmode div = -1;
  int newclock = TM->clock;

  switch(TM->request.graphtype)
   {
     case GT_TEXT:	 div = DAC_MUX_NORMAL;  break;
     case GT_1BIT:	 div = DAC_MUX_32; break;
     case GT_4BIT_PLANES:div = DAC_MUX_8;  break;
     case GT_8BIT:	 div = DAC_MUX_4;  break;

     case GT_15BIT:
     case GT_16BIT:	 div = DAC_MUX_2;  break;

     case GT_24BIT:      /* same as 32 bpp ??? */
     case GT_32BIT:	 div = DAC_MUX_1;  break;

     default:		printk("%s: could not handle graphtype %i\n",
		     __FILE__, TM->request.graphtype);
	return -ENOMODESUP_CHIP;
   }

#if 0
  if (newclock < ramdac_info.clock[div].min)
    {
      printk("%s: invalid clock (%i).\n", __FILE__, newclock);
      return -ENOMODESUP_CHIP;
    }

  if (ramdac_info.clock[div].max < newclock)	/* meet limits */
    newclock = ramdac_info.clock[div].max;

  if (ramdac_info.clock[DAC_MUX_NORMAL].max < newclock)
    {
      printk("%s: clock doubling mode not implemented yet.\n",__FILE__);
      newclock = ramdac_info.clock[DAC_MUX_NORMAL].max;
    }
#endif

  TM->clock    = newclock;
  TM->muxmode  = div;

  return MR_OK;
}

int kgi_ramdac_set_mode(struct kgi_mode * MT)
{ 
  int cr0=0;
  int cr1=0;
  int cr2=0;
  int cr3=0;

  cr0|=0x82;	/* We use 8 bit mode and want CR3 to be available */
  		/* FIXME: syncing options not implemented */
  
  switch(MT->muxmode)
  { case DAC_MUX_32:cr1|=0x80;break;
    case DAC_MUX_8 :cr1|=0x60;break;
    case DAC_MUX_4 :cr1|=0x40;break;
    case DAC_MUX_2 :switch(MT->request.graphtype)
    		    { case GT_15BIT:cr1|=0x30;break;
		      case GT_16BIT:cr1|=0x38;break;
		      default      :printk("Muxmode 2 for !=15/16 BPP.\n");return -ENOMODESUP_CHIP;
		    }
    case DAC_MUX_1 :cr1|=0x00;break;
    default	   :break;
  }
  
  /* We could use |0x10 in True-color modes to bypass CLUT, but the BT485 can do
     gamma-correction this way - we will use this feature. It has some effect on the
     Hicol modes as well, but I don't understand it, so I leave it disabled. */


  if (MT->muxmode!=DAC_MUX_NORMAL) cr2|=0x30;

  outbDAC(cr0,BT_COMMAND_REG_0);
  outbDAC(cr1,BT_COMMAND_REG_1);
  outbDAC(cr2,BT_COMMAND_REG_2);
  outbDAC(cr3,BT_STATUS_REG   ); /* This _is_ OK. */

  switch(MT->request.graphtype)
  {  case GT_TEXT:
     case GT_1BIT:
     case GT_4BIT_PLANES:
     case GT_8BIT:
     	hw_setpal(0,256,VGAPalette);break;

     case GT_15BIT:
     case GT_16BIT:
     case GT_24BIT:
     case GT_32BIT:
	hw_setpal(0,256,GammaMap);break;
  }
  hw_getpal(0,256,CurrState->palette);

  return MR_OK; }

static int hwc_off(void)
{ outbDAC(inbDAC(BT_COMMAND_REG_2)&0xfc,BT_COMMAND_REG_2);
  return(0); }

static int hwc_on(void)
{ int x,y;
  
  memcpy_fromfs(&ci, *(void **) (parameters+1), sizeof(struct ggi_cursorinfo));

  if (ci.sx<0||ci.sy<0) return -EINVAL;

  outbDAC(1,BT_CURS_WR_ADDR);
  outbDAC((ci.col1>>16)&0xff,BT_CURS_DATA);
  outbDAC((ci.col1>> 8)&0xff,BT_CURS_DATA);
  outbDAC((ci.col1    )&0xff,BT_CURS_DATA);
  outbDAC((ci.col2>>16)&0xff,BT_CURS_DATA);
  outbDAC((ci.col2>> 8)&0xff,BT_CURS_DATA);
  outbDAC((ci.col2    )&0xff,BT_CURS_DATA);

  outbDAC(inbDAC(BT_COMMAND_REG_2)|0x03,BT_COMMAND_REG_2);
  memcpy_tofs  (*(void **) (parameters+1),&ci , sizeof(struct ggi_cursorinfo));

  return(0); }

static int hwc_repos(void)
{ return 0; }

int kgi_ramdac_ioctl(struct inode *inode, struct file *file, \
        unsigned int cmd, unsigned long arg)
{ switch(cmd)
  { 
    case IO_RAMDAC_SETPAL :
    case IO_RAMDAC_SETGAMMA :
  	COPYPAR(2*sizeof(int)+sizeof(void *));
	return(setpal());

    case IO_RAMDAC_GETPAL :
    case IO_RAMDAC_GETGAMMA :
    	COPYPAR(2*sizeof(int)+sizeof(void *));
        return(getpal());

    case IO_RAMDAC_SETHWCDATA:
	COPYPAR(sizeof(int));
	if (*parameters)
	{ COPYPAR(sizeof(int)+sizeof(struct ggi_Cursorinfo *));
	  return hwc_on();
	} 
	else return hwc_off();

    case IO_RAMDAC_SETHWCPOS:
	COPYPAR(2*sizeof(int));
	return hwc_repos();
    		
    default : return -ENODRVSUP_ALWAYS_CANT; }
}

void kgi_ramdac_reset(void)
{ outbDAC(0x80, BT_COMMAND_REG_0);
  outbDAC(0x00, BT_STATUS_REG   );
  outbDAC(0x00, BT_COMMAND_REG_0);
  outbDAC(0x00, BT_COMMAND_REG_1);
  outbDAC(0x00, BT_COMMAND_REG_2);
}

int kgi_ramdac_init(void)
{ printk("Generic VGA Ramdac driver V"VERSION" loading ...\n");

  outbDAC(0x00,BT_COMMAND_REG_0 );
  if (0x88!=inbDAC(BT_STATUS_REG))
    printk("Status not 0x88 - this is probably not a BT485!\n");

  /* Set 6/8 bit mode and sync-on-green if required */
  outbDAC(0x02,BT_COMMAND_REG_0);

  hw_getpal(0,256,OpenState.palette);			

  return(0);
}

void kgi_ramdac_done(void) 
{ /* Back to 6 bit mode */
  hw_setpal(0,256,OpenState.palette);
  kgi_ramdac_reset();
}

#if 0

void 
s3BtCursorOn()
{
   unsigned char tmp;

   UNLOCK_SYS_REGS;

   /* turn on external cursor */
   outb(vgaCRIndex, 0x55);
   tmp = inb(vgaCRReg) & 0xDF;
   outb(vgaCRReg, tmp | 0x20);

   /* Enable Bt485 */
   if(!S3_x64_SERIES(s3ChipId)) {
     outb(vgaCRIndex, 0x45);
     tmp = inb(vgaCRReg) & 0xDF;
     outb(vgaCRReg, tmp | 0x20);
   }
   
   /* Enable cursor mode 3 - X11 mode */
   s3OutBtReg(BT_COMMAND_REG_2, 0xFC, 0x03);

   LOCK_SYS_REGS;
   return;
}

void
s3BtMoveCursor(pScr, x, y)
     ScreenPtr pScr;
     int   x, y;
{
   extern int s3hotX, s3hotY;
   extern Bool s3PixelMultiplexing;
   extern int s3AdjustCursorXPos;

   if (!xf86VTSema)
      return;

   if (s3BlockCursor)
      return;
   
   x -= s3InfoRec.frameX0 - s3AdjustCursorXPos;
   x += 64;
   x -= s3hotX;
   /* Compensate for using Bt485 Cursor without pixel multiplexing. */
   if (OFLG_ISSET(OPTION_STB_PEGASUS, &s3InfoRec.options) &&
       !s3PixelMultiplexing)
     x -= 2;

   if (x < 0)
      return;

   y -= s3InfoRec.frameY0;
   y += 64;
   y -= s3hotY;
   if (y < 0)
      return;

   UNLOCK_SYS_REGS;

   /* Output position - "only" 12 bits of location documented */
   s3OutBtReg(BT_CURS_X_LOW, 0x00, x & 0xFF);
   s3OutBtReg(BT_CURS_X_HIGH, 0x00, (x >> 8) & 0x0F);
   s3OutBtReg(BT_CURS_Y_LOW, 0x00, y & s3BtYPosMask);
   s3OutBtReg(BT_CURS_Y_HIGH, 0x00, (y >> 8) & 0x0F);

   LOCK_SYS_REGS;
   return;
}

void 
s3BtLoadCursor(pScr, pCurs, x, y)
     ScreenPtr pScr;
     CursorPtr pCurs;
     int x, y;
{
   int   index = pScr->myNum;
   register int   i, j;
   unsigned char *ram, *p, tmpcurs;
   extern int s3InitCursorFlag;

   if (!xf86VTSema)
      return;

   if (!pCurs)
      return;

   /* turn the cursor off */
   if ((tmpcurs = s3InBtReg(BT_COMMAND_REG_2)) & 0x03)
      s3BtCursorOff();

   /* load colormap */
   s3BtRecolorCursor(pScr, pCurs);

   ram = (unsigned char *)pCurs->bits->devPriv[index];

   UNLOCK_SYS_REGS;
   BLOCK_CURSOR;

   /* 
    * Bit 2 == 1 ==> 64x64 cursor; 2 low-order bits are extensions to the
    * address register, and must be cleared since we're going to start
    * writing data at address 0.
    */
   s3OutBtRegCom3(0xF8, 0x04);

   if (s3InfoRec.modes->Flags & V_INTERLACE) {
      s3StartBtData(BT_WRITE_ADDR, 0x00, BT_CURS_RAM_DATA);

      for (i=0; i < 64; i++) {		/* 64 rows in cursor */
	 p = ram + (((i % 2) ? i-1 : i+1)*8);
	 for (j=0; j < 8; j++,p++) {	/* 8 bytes per row */
            s3OutBtData(BT_CURS_RAM_DATA, *p);
	 }
      }
      ram += 512;
      for (i=0; i < 64; i++) {		/* 64 rows in cursor */
	 p = ram + (((i % 2) ? i-1 : i+1)*8);
	 for (j=0; j < 8; j++,p++) {	/* 8 bytes per row */
            s3OutBtData(BT_CURS_RAM_DATA, *p);
	 }
      }

      s3EndBtData();

      s3BtYPosMask = 0xFE;
      s3OutBtReg(BT_COMMAND_REG_2, 0xF7, 0x08);
   } else {
      /* Start data output */
      s3StartBtData(BT_WRITE_ADDR, 0x00, BT_CURS_RAM_DATA);

      /* 
       * Output the cursor data.  The realize function has put the planes into
       * their correct order, so we can just blast this out.
       */
      p = ram;
      for (i = 0; i < 1024; i++,p++)
         s3OutBtData(BT_CURS_RAM_DATA, *p);

      s3EndBtData();

      s3BtYPosMask = 0xFF;
      s3OutBtReg(BT_COMMAND_REG_2, 0xF7, 0x00);
   }

   UNBLOCK_CURSOR;
   LOCK_SYS_REGS;

   /* position cursor */
   s3BtMoveCursor(0, x, y);

   /* turn the cursor on */
   if ((tmpcurs & 0x03) || s3InitCursorFlag)
      s3BtCursorOn();

   if (s3InitCursorFlag)
      s3InitCursorFlag = FALSE;

   return;
}
#endif