/* 
   Graphics library for GGI. Horizontal lines.

   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define _LIB_COMPILE_

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <signal.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/time.h>
#include <linux/vt.h>
#include <linux/kd.h>
#include <string.h>
#include "../include/graphlib.h"


/**********************************/
/* draw/get/put a horizontal line */
/**********************************/

static int drawhline4bit(int unused,void *xyw)
{ 
	int adr;
	struct { int x,y,w; } *par;

	par=xyw;

	if ((par->x)&7)
	{	while(((par->x)&7)&&par->w) 
		{	drawpixel4bit(IO_GRAPH_DRAWPIXEL,xyw);
			par->x++;par->w--;
		}
	} 
	else	
		set16col(ggi_drwinfo.fg_color);

	adr=(par->y*CurrentMode.virtual.x+par->x)>>3;

	while(par->w>7)
	{	*(ggi_membuffer+adr)=0xff;
		adr++;par->w-=8;
	}

	while(par->w--)
	{	drawpixel4bit(IO_GRAPH_DRAWPIXEL,xyw);
		par->x++;
	}
	return 0;
}

static int drawhline8bit(int unused,void *xyw)
{ 
	struct { int x,y,w; } *par;

	par=xyw;
	memset(ggi_membuffer+par->y*CurrentMode.virtual.x+par->x,
	       ggi_drwinfo.fg_color,par->w);
  return 0;
}

static int drawhline_lower(int unused,void *xyw)
{ 
	struct { int x,y,w; } *par;

	par=xyw;
	while(par->w--) 
		ggi_drawpixel(par->x++,par->y);
	return(0);
}

int ggi_drawhline(int x,int y,int width)
{ 
	switchavail(IO_GRAPH_DRAWHLINE,&x);

	switcha(IO_GRAPH_DRAWHLINE)
	{ case AVM_MMAP :
		switch(CurrentMode.graphtype)
  		{ case GT_8BIT		:setdoemu(IO_GRAPH_DRAWHLINE,AVE_MODE,drawhline8bit,&x);
  		  case GT_4BIT_PLANES	:setdoemu(IO_GRAPH_DRAWHLINE,AVE_MODE,drawhline4bit,&x);
  		  default		:break;/* Use lower - no others now */
		}
	  case AVM_LOWER:
	  	setdoemu(IO_GRAPH_DRAWHLINE,AVE_ALWAYS,drawhline_lower,&x);
   		break;
	  default       :
	  	ggi_panic("hline - no method.");
	  	break;
  }
  return(0);
}

static int puthline8bit(int unused,void *xyw)
{ 
	struct { int x,y,w;
		 char *buffer; } *par;
	char *mem;

	par=xyw;
#if 0
	memcpy(ggi_membuffer+par->y*CurrentMode.virtual.x+par->x,
		par->buffer,par->w);
#else
	mem=ggi_membuffer+par->y*CurrentMode.virtual.x+par->x;
	while(par->w--)
	{ *mem++=*par->buffer++; }
#endif
	
	return 0;
}

static int puthline_lower(int unused,void *xyw)
{	
	struct { int x,y,w;
		 char *buffer; } *par;
	
	par=xyw;
  	while(par->w--) 
  		ggi_putpixel(par->x++,par->y,par->buffer++);
	return 0;
}

int ggi_puthline(int x,int y,int width,void *buffer)
{ 
	switchavail(IO_GRAPH_PUTHLINE,&x);

	switcha(IO_GRAPH_PUTHLINE)
	{ case AVM_MMAP :
		switch(CurrentMode.graphtype)
  		{ case GT_8BIT:setdoemu(IO_GRAPH_PUTHLINE,AVE_MODE,puthline8bit,&x);
	  	  default:break;
	  	}
	  case AVM_LOWER:setdoemu(IO_GRAPH_PUTHLINE,AVE_ALWAYS,puthline_lower,&x);
	  default       :
	  	ggi_panic("hline - no method.");break;
	}
	return(0);
}

static int gethline8bit(int unused,void *xyw)
{ 
	struct { int x,y,w;
		 char *buffer; } *par;

	par=xyw;
	memcpy(par->buffer,ggi_membuffer+par->y*CurrentMode.virtual.x+par->x,
		par->w);

	return 0;
}

static int gethline_lower(int unused,void *xyw)
{	
	struct { int x,y,w;
		 char *buffer; } *par;
	
	par=xyw;
  	while(par->w--) 
  		ggi_getpixel(par->x++,par->y,par->buffer++);
	return 0;
}

int ggi_gethline(int x,int y,int width,void *buffer)
{
	switchavail(IO_GRAPH_GETHLINE,&x);

	switcha(IO_GRAPH_GETHLINE)
	{ case AVM_MMAP :
		switch(CurrentMode.graphtype)
		{ case GT_8BIT:setdoemu(IO_GRAPH_GETHLINE,AVE_MODE,gethline8bit,&x);
  		  default: break;
  		}
	  case AVM_LOWER:setdoemu(IO_GRAPH_GETHLINE,AVE_ALWAYS,gethline_lower,&x);
	  default       :
	  	ggi_panic("hline - no method.");break;
	}
	return(0);
}

