/* 
   Graphics library for GGI. Pixels.

   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define _LIB_COMPILE_

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <signal.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/time.h>
#include <linux/vt.h>
#include <linux/kd.h>
#include <string.h>
#include "../include/graphlib.h"


/*******************************/
/* draw/get/put a single pixel */
/*******************************/

static int drawpixel32bit(int unused,void *xy)
{
	struct {int x,y; } *p;
	
	p=xy;
	
	*((long *)ggi_membuffer+p->y*CurrentMode.virtual.x+p->x)=ggi_drwinfo.fg_color;

	return(0); }

static int drawpixel16bit(int unused,void *xy)
{
	struct {int x,y; } *p;
	
	p=xy;
	
	*((short *)ggi_membuffer+p->y*CurrentMode.virtual.x+p->x)=ggi_drwinfo.fg_color;

	return(0); }

static int drawpixel8bit(int unused,void *xy)
{ 
	struct {int x,y; } *p;
	
	p=xy;
	
	*(ggi_membuffer+p->y*CurrentMode.virtual.x+p->x)=ggi_drwinfo.fg_color;

	return(0); }

static volatile char xyz;
static int fgcol=-1;

void set16col(int col)
{   if (fgcol!=col)
    { ggi_ioctl(IO_GRAPH_SET16COL,&col);
      fgcol=col; }
}

int drawpixel4bit(int unused,void *xy)
{ 	char msk;
	int adr;
	struct {int x,y; } *p;
	
	p=xy;
  
	set16col(ggi_drwinfo.fg_color);

	msk=0x80>>(p->x&7);adr=((p->y*CurrentMode.virtual.x)+p->x)>>3;

	xyz=*(ggi_membuffer+adr);
	*(ggi_membuffer+adr)=msk;
	return(0);
}

static int drawpixel_text(int unused,void *xy)
{ 
	struct {int x,y; } *p;
	
	p=xy;

	*(int *)(ggi_membuffer+p->y*CurrentMode.virtual.x+p->x*2)=
		0xdb+(ggi_drwinfo.fg_color<<8);
	return(0);
}

int ggi_drawpixel(int x,int y)
{ 
  	switchavail(IO_GRAPH_DRAWPIXEL,&x);

	switch(CurrentMode.graphtype)
	{ case GT_32BIT      :setdoemu(IO_GRAPH_DRAWPIXEL,AVE_MODE,drawpixel32bit,&x);break;
	  case GT_16BIT      :
	  case GT_15BIT      :setdoemu(IO_GRAPH_DRAWPIXEL,AVE_MODE,drawpixel16bit,&x);break;
	  case GT_8BIT       :setdoemu(IO_GRAPH_DRAWPIXEL,AVE_MODE,drawpixel8bit ,&x);break;
	  case GT_4BIT_PLANES:setdoemu(IO_GRAPH_DRAWPIXEL,AVE_MODE,drawpixel4bit ,&x);break;
	  case GT_TEXT       :setdoemu(IO_GRAPH_DRAWPIXEL,AVE_MODE,drawpixel_text,&x);break;
	  default	     :ggi_panic("drawpixel - no method");
	}
	return 0;
}

int ggi_getpixel(int x,int y,void *buffer)
{ switchavail(IO_GRAPH_GETPIXEL,&x);
  /* FIXME ! */
  *(char *)buffer=*(ggi_membuffer+y*CurrentMode.virtual.x+x);
  return 0;}


static int putpixel32bit(int unused,void *xy)
{
	struct {int x,y,*c; } *p;
	
	p=xy;
	
	*((long *)ggi_membuffer+p->y*CurrentMode.virtual.x+p->x)=*(p->c);

	return(0); }

static int putpixel16bit(int unused,void *xy)
{
	struct {int x,y;
	        short *c; } *p;
	
	p=xy;
	
	*((short *)ggi_membuffer+p->y*CurrentMode.virtual.x+p->x)=*(p->c);

	return(0); }

static int putpixel8bit(int unused,void *xy)
{ 
	struct {int x,y;
		char *c; } *p;
	
	p=xy;
	
	*(ggi_membuffer+p->y*CurrentMode.virtual.x+p->x)=*(p->c);

	return(0); }

int putpixel4bit(int unused,void *xy)
{ 	char msk;
	int adr;
	struct {int x,y;
		char *c; } *p;
	
	p=xy;
  
	set16col(*(p->c));

	msk=0x80>>(p->x&7);adr=((p->y*CurrentMode.virtual.x)+p->x)>>3;

	xyz=*(ggi_membuffer+adr);
	*(ggi_membuffer+adr)=msk;
	return(0);
}

static int putpixel_text(int unused,void *xy)
{ 
	struct {int x,y;
		char *c; } *p;
	
	p=xy;

	*(int *)(ggi_membuffer+p->y*CurrentMode.virtual.x+p->x*2)=
		0xdb+((*(p->c)&15)<<8);
	return(0);
}

int ggi_putpixel(int x,int y,void *buffer)
{
	switchavail(IO_GRAPH_PUTPIXEL,&x);

	switch(CurrentMode.graphtype)
	{ case GT_32BIT      :setdoemu(IO_GRAPH_PUTPIXEL,AVE_MODE,putpixel32bit,&x);break;
	  case GT_16BIT      :
	  case GT_15BIT      :setdoemu(IO_GRAPH_PUTPIXEL,AVE_MODE,putpixel16bit,&x);break;
	  case GT_8BIT       :setdoemu(IO_GRAPH_PUTPIXEL,AVE_MODE,putpixel8bit ,&x);break;
	  case GT_4BIT_PLANES:setdoemu(IO_GRAPH_PUTPIXEL,AVE_MODE,putpixel4bit ,&x);break;
	  case GT_TEXT       :setdoemu(IO_GRAPH_PUTPIXEL,AVE_MODE,putpixel_text,&x);break;
	  default	     :ggi_panic("putpixel - no method");
        }
	return 0;
}

