/* 
   Graphics library for GGI.

   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define _LIB_COMPILE_

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <signal.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/time.h>
#include <linux/vt.h>
#include <linux/kd.h>
#include <string.h>
#include "../include/graphlib.h"

int ggi_tty_fd=0;
int gd_shmid=-1;			/* id-number of shared memory     */
char *gd_shmem=NULL;			/* address of 1st shmem-attach (not
					   the shadowed one under ggi_membuffer */

/* Virtual console switching and memory locking */
static int forbidvtacquire = 0;
static int forbidvtrelease = 0;

/* This should NOT BE USED ! */
int lock_count = 0;
volatile int release_flag = 0;
volatile int acquire_flag = 0;

int ggi_vt_num;
static int sleep_when_switched_away=0;

/**************************/
/* get the controlling vt */
/**************************/
int ggi_getvt(int *vtnum)
{ return(ggi_ioctl(IO_DRIVER_GETVT,&vtnum)); }

/**********************************/
/* activate a VT - 0 means our VT */
/**********************************/
int ggi_vt_activate(int num)
{ if (num==0) num=ggi_vt_num;
  if (ioctl(ggi_tty_fd, VT_ACTIVATE  , num)||
      ioctl(ggi_tty_fd, VT_WAITACTIVE, num)  ) return -EIO;
  return(0); }

/***************************/
/* get currently active VT */
/***************************/
int ggi_get_active_vt(void)
{ struct vt_stat vtstat;
  if (ioctl(ggi_tty_fd, VT_GETSTATE , &vtstat)) return -EIO;
  return(vtstat.v_active); }

/***********************************/
/* set the switched-away behaviour */
/***********************************/
int ggi_run_in_bg(int yesno)
{ sleep_when_switched_away=!yesno;return(0); }

int ggi_bgmode(void)
{ return(ggi_ioctl(IO_DRIVER_BGMODE,NULL)); }

int ggi_fgmode(void)
{ return(ggi_ioctl(IO_DRIVER_FGMODE,NULL)); }

/*****************************/
/* Virtual console switching */
/*****************************/

void releasevt_signal( int n ) {
	struct sigaction siga;
	struct ggi_event ev;

	if (lock_count) {		/* Not yet used,but might come in handy */
		release_flag = 1;
		return;
	}
	forbidvtacquire = 1;
	SETSIG2(siga, SIGUSR1, releasevt_signal,old_handlers[SIGUSR1]);
	if (forbidvtrelease) {
		forbidvtacquire = 0;
		ioctl(ggi_tty_fd, VT_RELDISP, 0);
		return;
	}

	/* Move screen contents to shared memory buffer. */
	ggi_getbox(0,0,CurrentMode.virtual.x,CurrentMode.virtual.y,gd_shmem);
  
	/* Now map the shared mem to the place where the VGA RAM was */
	munmap((caddr_t)ggi_membuffer,MMAP_SIZE);
	shmat(gd_shmid,ggi_membuffer,0);

	ggi_bgmode();

	discard_avail();		/* Background - acceleration off */
	ioctl(ggi_tty_fd, VT_RELDISP, 1);
	forbidvtacquire = 0;

	ev.eventtype=EV_VT;
	ev.evdata.vt.subtype=VT_SWITCH_AWAY;
	ev.evdata.vt.console=0;		/* Don't know this here - is there a way short of
					   using ioctl GETSTATE ? */
	ggi_send_event(ev);

	if (sleep_when_switched_away)
	   ioctl(ggi_tty_fd, VT_WAITACTIVE, ggi_vt_num);
} 

void acquirevt_signal( int n ) {
	struct sigaction siga;
	struct ggi_event ev;

	if (lock_count) {		/* Not yet used,but might come in handy */
		acquire_flag = 1;
		return;
	}
	forbidvtrelease = 1;
	SETSIG2(siga, SIGUSR2, acquirevt_signal, old_handlers[SIGUSR2]);
	if (forbidvtacquire) {
		forbidvtrelease = 0;
		return;
	}

	ggi_fgmode();			/* start doing real work again */
	/* Remap the VGA RAM to the original address */
        shmdt(ggi_membuffer);
        mmap((caddr_t)ggi_membuffer,
		 MMAP_SIZE,              /* 2048K - FIXME ! */
		 PROT_READ|PROT_WRITE,   /* for reading and writing  */
		 MAP_SHARED,             /* shared mapping */
                 gd_hand,                /* from the graphics-device */
                 0x0                     /* from the very beginning  */
                 );

	discard_avail();		/* Foreground - acceleration on */
	/* Get screen contents from shared memory buffer. */
	ggi_putbox(0,0,CurrentMode.virtual.x,CurrentMode.virtual.y,gd_shmem);

	ioctl(ggi_tty_fd, VT_RELDISP, VT_ACKACQ);
	forbidvtrelease = 0;

	ev.eventtype=EV_VT;
	ev.evdata.vt.subtype=VT_SWITCH_TO;
	ev.evdata.vt.console=0;		/* Don't know this here - is there a way ? */
	ggi_send_event(ev);
}

void takevtcontrol(void) {
    	struct sigaction siga;
	struct vt_mode newvtmode;
	struct vt_mode oldvtmode;

	ggi_vt_activate(0);		/* Must be active on startup anything else
        				   will confuse the vt-switching-code      */
	ioctl(ggi_tty_fd, VT_GETMODE, &oldvtmode);
    	newvtmode = oldvtmode;
    	newvtmode.mode = VT_PROCESS;	/* handle VT changes */
    	newvtmode.relsig = SIGUSR1;	/* I didn't find SIGUSR1/2 anywhere */
    	newvtmode.acqsig = SIGUSR2;	/* in the kernel sources, so I guess */
    					/* they are free */
	SETSIG2(siga, SIGUSR1, releasevt_signal,old_handlers[SIGUSR1]);
    	SETSIG2(siga, SIGUSR2, acquirevt_signal,old_handlers[SIGUSR2]);
    	ioctl(ggi_tty_fd, VT_SETMODE, &newvtmode);
}

